#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import logging
import json
from collections import defaultdict
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import noload, joinedload

from ..util import hdb
from ..exceptions import PermissionError, HydraError
from ..db.model import Project, ProjectOwner, Network, NetworkOwner, User, Attr
from .. import db
from . import network
from .objects import JSONObject
from ..util.permissions import required_perms
from . import scenario
from ..exceptions import ResourceNotFoundError
from datetime import datetime

log = logging.getLogger(__name__)

def _get_project(project_id, user_id, check_write=False):
    try:
        project_i = db.DBSession.query(Project).filter(Project.id == project_id).options(noload(Project.children)).one()

        if check_write is True:
            project_i.check_write_permission(user_id)
        else:
            ## to avoid doing 2 checks, only check this if the check write is not set
            project_i.check_read_permission(user_id)

        return project_i
    except NoResultFound:
        raise ResourceNotFoundError("Project %s not found"%(project_id))

def _add_project_attribute_data(project_i, attr_map, attribute_data):
    if attribute_data is None:
        return []
    #As projects do not have scenarios (or to be more precise, they can only use
    #scenario 1, we can put
    #resource scenarios directly into the 'attributes' attribute
    #meaning we can add the data directly here.
    resource_scenarios = []
    for attr in attribute_data:
        if attr.resource_attr_id < 0:
            ra_i = attr_map[attr.resource_attr_id]
            attr.resource_attr_id = ra_i.id

        rscen = scenario._update_resourcescenario(None, attr)

        resource_scenarios.append(rscen)

    return resource_scenarios

@required_perms('add_project')
def add_project(project, **kwargs):
    """
        Add a new project
        returns a project complexmodel
    """
    user_id = kwargs.get('user_id')

    existing_proj = get_project_by_name(project.name, user_id=user_id)

    if len(existing_proj) > 0:
        if existing_proj[0].status == 'X':
            now = datetime.now().strftime("%Y%m%d%H%M%S")
            existing_proj[0].name = "{new_project_name} {now}"
            log.info("Updating an existing deleted project %s with new project name to avoid naming clash %s",
                      existing_proj[0].id, existing_proj[0].name)
        else:
            raise HydraError(f'A Project with the name "{project.name}" already exists')

    proj_i = Project()

    #'appdata' is a metadata column. It's not called 'metadata' because
    #'metadata' is a reserved sqlalchemy keyword.
    #Add appdata to the core columns for insertion done like this because
    #updating is done differently (where the contents are updated, not the whole column)
    for columnname in Project.core_columns + ['appdata']:
        if column := getattr(project, columnname, None):
            setattr(proj_i, columnname, column)

    proj_i.created_by = user_id

    #A project can only be added to another if the user has write access to the target,
    #so we need to check the permissions on the target project if it is specified
    if parent_id := getattr(project, 'parent_id', None):
        #check the user has the correct permission to write to the target project
        _get_project(parent_id, user_id, check_write=True)
        proj_i.parent_id = parent_id

    attr_map = hdb.add_resource_attributes(proj_i, project.attributes)

    db.DBSession.flush() #Needed to get the resource attr's ID

    proj_data = _add_project_attribute_data(proj_i, attr_map, project.attribute_data)

    proj_i.attribute_data = proj_data

    proj_i.set_owner(user_id)

    db.DBSession.add(proj_i)
    db.DBSession.flush()

    Project.clear_cache(user_id)

    return proj_i

@required_perms('edit_project')
def update_project(project, **kwargs):
    """
        Update a project
        returns a project complexmodel
    """

    user_id = kwargs.get('user_id')

    proj_i = _get_project(project.id, user_id, check_write=True)

    for columnname in Project.core_columns:
        if column := getattr(project, columnname, None):
            setattr(proj_i, columnname, column)

    #'appdata' is a metadata column. It's not called 'metadata' because
    #'metadata' is a reserved sqlalchemy keyword.
    #Rather than replace the column
    if appdata := getattr(project, 'appdata', None):
        if proj_i.appdata is None:
            proj_i.appdata = project.appdata
        else:
            newdict = proj_i.appdata.copy()
            newdict.update(project.appdata)
            proj_i.appdata = newdict

    #A project can only be moved to another if the user has write access on both,
    #so we need to check the permissions on the target project if it is specified
    if parent_id := getattr(project, "parent_id", None):
        if parent_id != proj_i.parent_id:
            #check the user has the correct permission to write to the target project
            _get_project(project.parent_id, user_id, check_write=True)
            proj_i.parent_id = project.parent_id
    else:
        # parent_id has changed to None
        if proj_i.parent_id is not None:
            proj_i.parent_id = None

    if project.attributes:
        attr_map = hdb.add_resource_attributes(proj_i, project.attributes)
        proj_data = _add_project_attribute_data(proj_i, attr_map, project.attribute_data)
        proj_i.attribute_data = proj_data

    Project.clear_cache(user_id)

    db.DBSession.flush()

    return proj_i

@required_perms('edit_project')
def move_project(project_id, target_project_id, **kwargs):
    """
        Move a project from one project into another
    """

    user_id = kwargs.get('user_id')

    #Check the user has access to write to the project
    proj_i = _get_project(project_id, user_id, check_write=True)

    if target_project_id is None:
        return remove_project_parent(project_id, **kwargs)

    #check the user has the correct permission to write to the target project
    _get_project(target_project_id, user_id, check_write=True)

    Project.clear_cache(user_id)

    proj_i.parent_id = target_project_id

    db.DBSession.flush()

    return proj_i

@required_perms('edit_project')
def remove_project_parent(project_id, **kwargs):
    """
        Removes the parent of the <project_id> argument
        by setting the parent_id attr to None
    """
    user_id = kwargs.get('user_id')
    proj_i = _get_project(project_id, user_id, check_write=True)
    Project.clear_cache(user_id)
    proj_i.parent_id = None
    db.DBSession.flush()

    return proj_i

@required_perms('get_project')
def check_project_write_permission(project_id, user_id):
    """
        Check whether a user has write permission on a project
        Args:
            project_id (int): ID of the project
            user_id (int): ID of the user
        Returns:
            True if the user has write permission
            False if the user does not have write permission
    """
    proj_i = _get_project(project_id, user_id)
    return proj_i.check_write_permission(user_id)

@required_perms('get_project')
def check_project_read_permission(project_id, user_id):
    """
        Check whether a user has read permission on a project
        Args:
            project_id (int): ID of the project
            user_id (int): ID of the user
        Returns:
            True if the user has read permission
            False if the user does not have read permission
    """
    proj_i = _get_project(project_id, user_id)
    return proj_i.check_read_permission(user_id)

@required_perms('get_project')
def get_project(project_id, include_deleted_networks=False, **kwargs):
    """
        Get a project object
        Args:
            project_id (int): The ID of the project
            include_deleted_networks (bool): Include networks with the status 'X'. False by default
        returns:
            JSONObject of the project
    """
    user_id = kwargs.get('user_id')
    log.info("Getting project %s", project_id)

    proj_i = _get_project(project_id, user_id)

    #lazy load owners
    proj_i.attributes#pylint: disable=W0104

    nav_only = False
    if not proj_i.is_owner(user_id):
        if user_id not in [o.user_id for o in proj_i.owners]:
            proj_i.owners = []
            proj_i.attributes = []
            nav_only = True


    proj_j = JSONObject(proj_i)

    proj_j.owners = proj_i.get_owners()

    proj_j.nav_only = nav_only

    proj_j.attribute_data = [JSONObject(rs) for rs in proj_i.get_attribute_data()]

    proj_j.networks = proj_i.get_networks(
        user_id,
        include_deleted_networks=include_deleted_networks)

    proj_j.projects = proj_i.get_child_projects(
        user_id,
        include_deleted_networks=include_deleted_networks)


    log.info("Project %s retrieved", project_id)

    return proj_j

@required_perms('get_project')
def get_project_by_network_id(network_id, **kwargs):
    """
        get a project object by a network_id
    """
    user_id = kwargs.get('user_id')

    project_i = db.DBSession.query(Project)\
        .join(Network, Project.id == Network.project_id)\
        .filter(Network.id == network_id)\
        .order_by('name').one()

    project_i.check_read_permission(user_id)

    return project_i



@required_perms('get_project')
def get_project_by_name(project_name, **kwargs):
    """
        get a project complexmodel
    """
    user_id = kwargs.get('user_id')

    projects_i = db.DBSession.query(Project).join(ProjectOwner).filter(
        Project.name == project_name,
        ProjectOwner.user_id == user_id,
        ProjectOwner.view == 'Y').order_by('name').all()

    return projects_i

@required_perms('get_project')
def get_projects(uid, include_shared_projects=True, projects_ids_list_filter=None, project_id=None, **kwargs):
    """
        Get all the projects owned by the specified user.
        These include projects created by the user, but also ones shared with the user.
        For shared projects, only include networks in those projects which are accessible to the user.

        the include_shared_projects flag indicates whether to include projects which have been shared
        with the user, or to only return projects created directly by this user.
    """
    req_user_id = kwargs.get('user_id')

    Project.build_user_cache(uid)

    ##Don't load the project's networks. Load them separately, as the networks
    #must be checked individually for ownership
    projects_qry = db.DBSession.query(Project).options(joinedload(Project.owners))

    log.info("Getting projects for user %s", uid)

    if project_id is not None:
        log.info("Getting projects in project %s", project_id)
    else:
        log.info("Getting top-level projects")

    if include_shared_projects is True:
        projects_qry = projects_qry.outerjoin(ProjectOwner).filter(
            Project.status == 'A',
            ProjectOwner.user_id == uid,
            ProjectOwner.view == 'Y')
    else:
        projects_qry = projects_qry.filter(Project.created_by == uid)

    if projects_ids_list_filter is not None:
        # Filtering the search of project id
        if isinstance(projects_ids_list_filter, str):
            # Trying to read a csv string
            projects_ids_list_filter = json.loads(projects_ids_list_filter)
            if isinstance(projects_ids_list_filter, int):
                projects_ids_list_filter = [projects_ids_list_filter]
                projects_qry = projects_qry.filter(Project.id.in_(projects_ids_list_filter))


    projects_qry = projects_qry.options(noload(Project.networks)).order_by('id')

    projects_i = projects_qry.all()

    log.info("Project query done for user %s. %s projects found", uid, len(projects_i))

    #now separate all the projects in the current scope (in the requested project ID)
    #from the ones that are not. Using the ones that are not, we need to find the projects
    #which allow the user access to the projects that they have prermissions on

    scoped_projects = [p for p in projects_i if p.parent_id == project_id]
    scoped_project_ids = {p.id for p in scoped_projects}

    #Now get projects which the user must have access to in order to navigate
    #to projects further down the tree which they are owners of.
    nav_project_ids = set(Project.get_cache(uid).get(project_id, [])) - scoped_project_ids
    nav_projects_i = db.DBSession.query(Project).filter(Project.id.in_(nav_project_ids)).filter(Project.parent_id==project_id).all()
    nav_projects = []
    for nav_project_i in nav_projects_i:
        nav_project_j = JSONObject(nav_project_i)
        nav_project_j.nav_only = True
        nav_project_j.owners = []
        nav_project_j.networks = []
        nav_projects.append(nav_project_j)


    user = db.DBSession.query(User).filter(User.id == req_user_id).one()
    isadmin = user.is_admin()

    project_network_lookup = get_projects_networks([p.id for p in projects_i], uid, isadmin=isadmin, **kwargs)

    #Load each
    projects_j = []
    for project_i in scoped_projects:
        project_i.attributes#pylint: disable=W0104
        project_i.get_attribute_data()
        project_j = JSONObject(project_i)
        project_j.networks = project_network_lookup.get(project_i.id, [])
        project_j.projects = [JSONObject(p) for p in project_i.get_child_projects(req_user_id)]
        projects_j.append(project_j)

    log.info("Networks loaded projects for user %s", uid)

    return projects_j + nav_projects

def get_projects_in(project_ids, **kwargs):
    """
        Get the list of projects specified in a list of IDS
        args:
            project_ids (list(int)) : a list of project IDs
        returns:
            list(Project)
    """
    user_id = kwargs.get('user_id')

    projects = db.DBSession.query(Project).filter(
            Project.id.in_(project_ids))

    for p in projects:
        p.check_read_permission(user_id)

    return projects


def get_projects_networks(project_ids, uid, isadmin=None, **kwargs):
    """
        Get all the networks in all the projects specified, checking for ownership
        with the user ID
        args:
            project_ids (list): a list of integer project IDs
            uid: the user ID for whom the request is being made so we can check for ownershiop
            isadmin: A flag to indicate if the requesting user is an admin. If null, does a query to find out
        returns:
            dict: A lookup keyed on project ID, with values being a list of networks
    """
    #Do a single query for all the networks in all the user's projects,
    #then make a lookup dictionary so that each projec's networks can be grouped
    #together, and accessed later.

    user_id = kwargs.get('user_id')
    if isadmin is None:
        req_user_id = kwargs.get('user_id')
        user = db.DBSession.query(User).filter(User.id == req_user_id).one()
        isadmin = user.is_admin()

    log.info("Getting for all the networks for in the specified projects...")
    network_qry = db.DBSession.query(Network)\
                                .options(joinedload(Network.owners))\
                                .filter(Network.project_id.in_(project_ids),\
                                        Network.status=='A')
    if not isadmin:
        network_qry.outerjoin(NetworkOwner)\
        .filter(
            NetworkOwner.user_id == uid,
            NetworkOwner.view == 'Y'
        )

    networks = network_qry.all()
    project_network_lookup = defaultdict(list)
    for network_i in networks:
        net_j = JSONObject(network_i)
        if net_j.layout is not None:
            net_j.layout = JSONObject(net_j.layout)
        else:
            net_j.layout = JSONObject({})
        project_network_lookup[net_j.project_id].append(net_j)

    log.debug("Network query done")

    return project_network_lookup


@required_perms('get_project')
def get_project_attribute_data(project_id, **kwargs):
    req_user_id = kwargs.get('user_id')

    project_i = _get_project(project_id, req_user_id)

    return project_i.get_attribute_data()

@required_perms('delete_project')
def set_project_status(project_id, status, **kwargs):
    """
        Set the status of a project to 'X' or 'A'
    """
    user_id = kwargs.get('user_id')
    project = _get_project(project_id, user_id, check_write=True)
    project.status = status
    if status.lower() == 'x':
        #delete the project from the cache
        project.remove_from_cache()
    db.DBSession.flush()

@required_perms('edit_project', 'delete_project')
def delete_project(project_id, **kwargs):
    """
        Set the status of a project to 'X'
    """
    user_id = kwargs.get('user_id')
    project = _get_project(project_id, user_id, check_write=True)
    db.DBSession.delete(project)
    db.DBSession.flush()

    return 'OK'

@required_perms("get_project")
def get_networks(project_id, include_data='N', **kwargs):
    """
        Get all networks in a project
        Returns an array of network objects.
    """
    log.info("Getting networks for project %s", project_id)
    user_id = kwargs.get('user_id')
    project = _get_project(project_id, user_id)

    rs = db.DBSession.query(Network.id, Network.status).filter(
        Network.project_id == project_id).all()

    networks = []
    for r in rs:
        if r.status != 'A':
            continue
        try:
            net = network.get_network(r.id, summary=True, include_data=include_data, **kwargs)
            log.info("Network %s retrieved", net.name)
            networks.append(net)
        except PermissionError:
            log.info("Not returning network %s as user %s does not have "
                     "permission to read it."%(r.id, user_id))

    return networks

@required_perms('get_project')
def get_network_project(network_id, **kwargs):
    """
        get the project that a network is in
    """

    net_proj = db.DBSession.query(Project)\
        .join(Network, Project.id == Network.project_id)\
        .filter(Network.id == network_id).first()

    if net_proj is None:
        raise HydraError("Network %s not found"% network_id)

    return net_proj

@required_perms('get_project', 'add_project')
def clone_project(project_id,
    recipient_user_id=None,
    new_project_name=None,
    new_project_description=None,
    creator_is_owner=False,
    **kwargs):
    """
        Create an exact clone of the specified project for the specified user.
        args:
            recipient_user_id (int): The ID of the user who will be granted ownership of the project after cloning.
                If None, ownership will be granted to the requesting user.
            new_project_name (str): The name of the cloned project. If None, then the project's name will be postfixed with ('Cloned by XXX')
            new_project_description (str): The description of the cloned project.
                If None, the current project's description will be used.
            creator_is_owner (Bool) : The user who creates the network isn't added as an owner
                (won't have an entry in tNetworkOwner and therefore won't see the network in 'get_project')
        returns:
            (int): The ID of the newly created project
    """

    user_id = kwargs['user_id']

    log.info("Creating a new project for cloned network")

    project = _get_project(project_id, user_id, check_write=True)
    if recipient_user_id is None:
        recipient_user_id = user_id

    recipient_user = db.DBSession.query(User).filter(User.id == recipient_user_id).one()
    cloning_user = db.DBSession.query(User).filter(User.id == user_id).one()

    if new_project_name is None:
        new_project_name = project.name + ' Cloned By {}'.format(cloning_user.display_name)

    #check a project with this name doesn't already exist:
    project_with_name =  db.DBSession.query(Project).filter(
        Project.name == new_project_name,
        Project.created_by == recipient_user_id).first()

    if project_with_name is not None:
        now = datetime.now().strftime("%Y%m%d%H%M%S")
        if project_with_name.status == 'X':
            project_with_name.name = f"{new_project_name} {now}"
            log.info("Updating an existing deleted project %s with new project anme to avoid naming clash %s",
                      project_with_name.id, project_with_name.name)
        else:
            #If there is a naming clash, then add a timestamp to the name of the new project to make it unique for
            #this user.
            new_project_name = f"{new_project_name} ({now})"

    new_project = Project()
    new_project.name = new_project_name
    if new_project_description:
        new_project.description = new_project_description
    else:
        new_project.description = project.description

    new_project.appdata = project.appdata

    new_project.created_by = recipient_user_id

    if recipient_user_id is not None and recipient_user_id != user_id:
        project.check_share_permission(user_id)
        new_project.set_owner(recipient_user_id)

    if creator_is_owner is True:
        new_project.set_owner(user_id)

    db.DBSession.add(new_project)
    db.DBSession.flush()

    network_ids = db.DBSession.query(Network.id).filter(
                                        Network.project_id==project_id).all()

    for n in network_ids:
        network.clone_network(n.id,
                              recipient_user_id=recipient_user_id,
                              project_id=new_project.id,
                              user_id=user_id)

    project_attributes = db.DBSession.query(Attr).filter(Attr.project_id==project_id).all()

    for pa in project_attributes:
        newpa = Attr()
        newpa.name = pa.name
        newpa.dimension_id = pa.dimension_id
        newpa.description = pa.description
        newpa.project_id = new_project.id
        db.DBSession.add(newpa)

    db.DBSession.flush()

    return new_project.id

def get_project_hierarchy(project_id, **kwargs):
    """
        Return a list of project JSONObjects which represent the links in the chain up to the root project
        [project_j, parent_j, parent_parent_j ...etc]
        If the project has no parent, return [project_j]
    """
    user_id = kwargs.get('user_id')
    hierarchy = _get_project(project_id, user_id).get_hierarchy(user_id)
    return hierarchy



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.orm.exc import NoResultFound

from .objects import JSONObject
from ..db.model import User, Role, Perm, RoleUser, RolePerm
from ..exceptions import ResourceNotFoundError, HydraError, HydraLoginUserNotFound
from .. import db
from .. import config
from ..util.permissions import required_perms

import bcrypt
import logging

log = logging.getLogger(__name__)

def _get_user_id(username):
    try:
        rs = db.DBSession.query(User.id).filter(User.username==username).one()
        return rs.id
    except:
        return None

def _get_user(user_id, **kwargs):
    try:
        user_i = db.DBSession.query(User).filter(User.id==user_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("User %s does not exist"%user_id)

    return user_i

def _get_role(role_id,**kwargs):
    try:
        role_i = db.DBSession.query(Role).filter(Role.id==role_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Role %s does not exist"%role_id)

    return role_i

def _get_perm(perm_id,**kwargs):
    try:
        perm_i = db.DBSession.query(Perm).filter(Perm.id==perm_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Permission %s does not exist"%perm_id)

    return perm_i

@required_perms("get_user")
def get_username(uid,**kwargs):
    """
        Return the username of a given user_id
    """
    rs = db.DBSession.query(User.username).filter(User.id==uid).one()

    if rs is None:
        raise ResourceNotFoundError("User with ID %s not found"%uid)

    return rs.username

@required_perms("get_user")
def get_usernames_like(username,**kwargs):
    """
        Return a list of usernames like the given string.
    """
    checkname = "%%%s%%"%username
    rs = db.DBSession.query(User.username).filter(User.username.like(checkname)).all()
    return [r.username for r in rs]


@required_perms("add_user")
def add_user(user, **kwargs):
    """
        Add a user
    """
    u = User()

    u.username     = user.username
    u.display_name = user.display_name

    user_id = _get_user_id(u.username)

    #If the user is already there, cannot add another with
    #the same username.
    if user_id is not None:
        raise HydraError("User %s already exists!"%user.username)

    u.password = bcrypt.hashpw(str(user.password).encode('utf-8'), bcrypt.gensalt())

    db.DBSession.add(u)
    db.DBSession.flush()

    return u

@required_perms("edit_user")
def update_user_display_name(user, **kwargs):
    """
        Update a user's display name
    """
    try:
        user_i = db.DBSession.query(User).filter(User.id == user.id).one()
        user_i.display_name = user.display_name
        return user_i
    except NoResultFound:
        raise ResourceNotFoundError("User (id=%s) not found"%(user.id))

@required_perms("edit_user")
def update_user_password(new_pwd_user_id, new_password, **kwargs):
    """
        Update a user's password
    """
    try:
        user_i = db.DBSession.query(User).filter(User.id == new_pwd_user_id).one()
        user_i.password = bcrypt.hashpw(str(new_password).encode('utf-8'), bcrypt.gensalt())
        reset_failed_logins(user_i.username, flush=False, **kwargs)
        db.DBSession.flush()
        return user_i
    except NoResultFound:
        raise ResourceNotFoundError("User (id=%s) not found"%(new_pwd_user_id))

@required_perms("get_user")
def get_user(uid, **kwargs):
    """
        Get a user by ID
    """
    user_id=kwargs.get('user_id')
    if uid is None:
        uid = user_id
    user_i = _get_user(uid)
    return user_i

@required_perms("get_user")
def get_user_by_name(uname, **kwargs):
    """
        Get a user by username
    """
    try:
        user_i = db.DBSession.query(User).filter(User.username == uname).one()
        return user_i
    except NoResultFound:
        return None

@required_perms("get_user")
def get_user_by_id(uid,**kwargs):
    """
        Get a user by username
    """
    user_id = kwargs.get('user_id')
    try:
        user_i = _get_user(uid)
        return user_i
    except NoResultFound:
        return None

@required_perms("edit_user")
def delete_user(deleted_user_id,**kwargs):
    """
        Delete a user
    """
    try:
        user_i = db.DBSession.query(User).filter(User.id==deleted_user_id).one()
        db.DBSession.delete(user_i)
    except NoResultFound:
        raise ResourceNotFoundError("User (user_id=%s) does not exist"%(deleted_user_id))


    return 'OK'

@required_perms("get_user")
def get_failed_login_count(username, **kwargs):
    try:
        user_i = db.DBSession.query(User).filter(User.username == username).one()
    except NoResultFound:
        """ Non-existent user should not raise """
        return -1

    if user_i.failed_logins is None:
        log.info("User failed login count is None. Defaulting to 0")
        return 0

    return user_i.failed_logins


@required_perms("get_user")
def get_max_login_attempts(*args, **kwargs):
    """
        Get the max number of login attempts a user can have. This usually
        comes from config.
        NOTE:
         In the absence of max_login_attempts defined in config,
         a value of 0 will be returned and users will be unable to log
         in.
    """
    max_login_attempts = int(config.get("security", "max_login_attempts", 0))

    return max_login_attempts


@required_perms("get_user")
def get_remaining_login_attempts(username, **kwargs):
    """
        Get a user's remaininig login attempts by subtracting the number
        of failed login attempts from the number of allowed login attempts
        args:
            username(str): The username to lookup the user
        returns:
            The remaining number of login attempots
        raises:
            HydraError if the username does not exist
    """
    max_login_attempts = get_max_login_attempts(**kwargs)
    failed_attempts = get_failed_login_count(username, **kwargs)

    log.info("Max login attempts: %s, Failed logins: %s", max_login_attempts, failed_attempts)

    return max_login_attempts - failed_attempts


@required_perms("edit_user")
def inc_failed_login_attempts(username, flush=True, **kwargs):
    """
        Increment a user's failed login count
        args:
            username(str): The username to lookup the user
            flush (bool, default=True):
                   A flag to indicate whether this should flush to the DB. This
                   exists as this function may be called as part of another function
                   which will do the flush later.
        returns:
            void
        raises:
            HydraError if the username does not exist
    """
    try:
        user_i = db.DBSession.query(User).filter(User.username == username).one()
    except NoResultFound:
        raise HydraLoginUserNotFound("User '{}' does not exist".format(username))

    if user_i.failed_logins is not None:
        user_i.failed_logins = user_i.failed_logins + 1
    else:
        user_i.failed_logins = 1

    log.info("[ inc_failed_login_attempts ] User {} has currently {} failed logins".format(username, user_i.failed_logins))
    if flush is True:
        db.DBSession.flush()


@required_perms("edit_user")
def reset_failed_logins(username, flush=True, **kwargs):
    """
        Reset a user's failed login count to 0
        args:
            username (str): The username to lookup the user
            flush (bool, default=True):
                   A flag to indicate whether this should flush to the DB. This
                   exists as this function may be called as part of another function
                   which will do the flush later.
        returns:
            void
        raises:
            HydraError if the username does not exist
    """
    try:
        user_i = db.DBSession.query(User).filter(User.username == username).one()
    except NoResultFound:
        # raise HydraError(username)
        raise HydraLoginUserNotFound("User '{}' does not exist".format(username))

    user_i.failed_logins = 0

    if flush is True:
        db.DBSession.flush()


@required_perms("add_role")
def add_role(role,**kwargs):
    """
        Add a new role
    """
    role_i = Role(name=role.name, code=role.code)
    db.DBSession.add(role_i)
    db.DBSession.flush()

    return role_i

@required_perms("edit_role")
def delete_role(role_id,**kwargs):
    """
        Delete a role
    """
    try:
        role_i = db.DBSession.query(Role).filter(Role.id==role_id).one()
        db.DBSession.delete(role_i)
    except InvalidRequestError:
        raise ResourceNotFoundError("Role (role_id=%s) does not exist"%(role_id))

    return 'OK'

@required_perms("add_perm")
def add_perm(perm,**kwargs):
    """
        Add a permission
    """
    perm_i = Perm(name=perm.name, code=perm.code)
    db.DBSession.add(perm_i)
    db.DBSession.flush()

    return perm_i

@required_perms("edit_perm")
def delete_perm(perm_id,**kwargs):
    """
        Delete a permission
    """

    try:
        perm_i = db.DBSession.query(Perm).filter(Perm.id==perm_id).one()
        db.DBSession.delete(perm_i)
    except InvalidRequestError:
        raise ResourceNotFoundError("Permission (id=%s) does not exist"%(perm_id))

    return 'OK'


@required_perms("edit_user")
def set_user_role(new_user_id, role_id, **kwargs):
    """
        Apply `role_id` to `new_user_id`

        Note this function returns the `Role` instance associated with `role_id`
    """
    try:
        _get_user(new_user_id)
        role_i = _get_role(role_id)
        roleuser_i = RoleUser(user_id=new_user_id, role_id=role_id)
        role_i.roleusers.append(roleuser_i)
        db.DBSession.flush()
    except Exception as e: # Will occur if the foreign keys do not exist
        log.exception(e)
        raise ResourceNotFoundError("User or Role does not exist")

    return role_i

@required_perms("edit_user")
def delete_user_role(deleted_user_id, role_id,**kwargs):
    """
        Remove a user from a role
    """
    try:
        _get_user(deleted_user_id)
        _get_role(role_id)
        roleuser_i = db.DBSession.query(RoleUser).filter(RoleUser.user_id==deleted_user_id, RoleUser.role_id==role_id).one()
        db.DBSession.delete(roleuser_i)
    except NoResultFound:
        raise ResourceNotFoundError("User Role does not exist")

    return 'OK'

@required_perms("edit_role")
def set_role_perm(role_id, perm_id,**kwargs):
    """
        Insert a permission into a role
    """

    _get_perm(perm_id)
    role_i = _get_role(role_id)
    roleperm_i = RolePerm(role_id=role_id, perm_id=perm_id)

    role_i.roleperms.append(roleperm_i)

    db.DBSession.flush()

    return role_i


@required_perms("edit_role")
def delete_role_perm(role_id, perm_id,**kwargs):
    """
        Remove a permission from a role
    """
    _get_perm(perm_id)
    _get_role(role_id)

    try:
        roleperm_i = db.DBSession.query(RolePerm).filter(RolePerm.role_id==role_id, RolePerm.perm_id==perm_id).one()
        db.DBSession.delete(roleperm_i)
    except NoResultFound:
        raise ResourceNotFoundError("Role Perm does not exist")

    return 'OK'

@required_perms("edit_role")
def update_role(role,**kwargs):
    """
        Update the role.
        Used to add permissions and users to a role.
    """
    try:
        role_i = db.DBSession.query(Role).filter(Role.id==role.id).one()
        role_i.name = role.name
        role_i.code = role.code
    except NoResultFound:
        raise ResourceNotFoundError("Role (role_id=%s) does not exist"%(role.id))

    for perm in role.permissions:
        _get_perm(perm.id)
        roleperm_i = RolePerm(role_id=role.id,
                              perm_id=perm.id
                              )

        db.DBSession.add(roleperm_i)

    for user in role.users:
        _get_user(user.id)
        roleuser_i = RoleUser(user_id=user.id,
                                         perm_id=perm.id
                                        )

        db.DBSession.add(roleuser_i)

    db.DBSession.flush()
    return role_i


def get_all_users(**kwargs):
    """
        Get the username & ID of all users.
        Use the the filter if it has been provided
        The filter has to be a list of values
    """
    users_qry = db.DBSession.query(User)

    filter_type = kwargs.get('filter_type')
    filter_value = kwargs.get('filter_value')

    if filter_type is not None:
        # Filtering the search of users
        if filter_type == "id":
            if isinstance(filter_value, str):
                # Trying to read a csv string
                log.info("[HB.users] Getting user by Filter ID : %s", filter_value)
                filter_value = eval(filter_value)
                if type(filter_value) is int:
                    users_qry = users_qry.filter(User.id == filter_value)
                else:
                    users_qry = users_qry.filter(User.id.in_(filter_value))
        elif filter_type == "username":
            if isinstance(filter_value, str):
                # Trying to read a csv string
                log.info("[HB.users] Getting user by Filter Username : %s", filter_value)
                filter_value = filter_value.split(",")
                for i, em in enumerate(filter_value):
                    log.info("[HB.users] >>> Getting user by single Username : %s", em)
                    filter_value[i] = em.strip()
                if isinstance(filter_value, str):
                    users_qry = users_qry.filter(User.username == filter_value)
                else:
                    users_qry = users_qry.filter(User.username.in_(filter_value))
        else:
            raise Exception("Filter type '{}' not allowed".format(filter_type))

    else:
        log.info('[HB.users] Getting All Users')

    rs = users_qry.all()

    #In a 'get all' function, we only want to return basic info, not passwords,
    #login counts etc.
    basic_user_data = []

    for user in rs:
        basic_user_data.append(JSONObject({
            'id': user.id,
            'username': user.username,
            'display_name': user.display_name
        }))

    return basic_user_data

def get_all_perms(**kwargs):
    """
        Get all permissions
    """
    rs = db.DBSession.query(Perm).all()
    return rs

def get_all_roles(**kwargs):
    """
        Get all roles
    """
    rs = db.DBSession.query(Role).all()
    return rs

def get_role(role_id,**kwargs):
    """
        Get a role by its ID.
    """
    try:
        role = db.DBSession.query(Role).filter(Role.id==role_id).one()
        return role
    except NoResultFound:
        raise HydraError("Role not found (role_id={})".format(role_id))

@required_perms("get_user")
def get_user_roles(uid,**kwargs):
    """
        Get the roles for a user.
        @param user_id
    """
    try:
        user_roles = db.DBSession.query(Role).filter(Role.id==RoleUser.role_id,
                                                  RoleUser.user_id==uid).all()
        return user_roles
    except NoResultFound:
        raise HydraError("Roles not found for user (user_id={})".format(uid))

@required_perms("get_user")
def get_user_permissions(uid, **kwargs):
    """
        Get the roles for a user.
        @param user_id
    """
    try:
        _get_user(uid)

        user_perms = db.DBSession.query(Perm).filter(Perm.id==RolePerm.perm_id,
                                                  RolePerm.role_id==Role.id,
                                                  Role.id==RoleUser.role_id,
                                                  RoleUser.user_id==uid).all()
        return user_perms
    except:
        raise HydraError("Permissions not found for user (user_id={})".format(uid))

@required_perms("get_role")
def get_role_by_code(role_code,**kwargs):
    """
        Get a role by its code
    """
    try:
        role = db.DBSession.query(Role).filter(Role.code==role_code).one()
        return role
    except NoResultFound:
        raise ResourceNotFoundError("Role not found (role_code={})".format(role_code))


@required_perms("get_perm")
def get_perm(perm_id,**kwargs):
    """
        Get all permissions
    """

    try:
        perm = db.DBSession.query(Perm).filter(Perm.id==perm_id).one()
        return perm
    except NoResultFound:
        raise ResourceNotFoundError("Permission not found (perm_id={})".format(perm_id))

@required_perms("get_perm")
def get_perm_by_code(perm_code,**kwargs):
    """
        Get a permission by its code
    """

    try:
        perm = db.DBSession.query(Perm).filter(Perm.code==perm_code).one()
        return perm
    except NoResultFound:
        raise ResourceNotFoundError("Permission not found (perm_code={})".format(perm_code))



# (c) Copyright 2013, 2014, University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import copy
import json
import hydra_base
import pytest
from hydra_base.exceptions import HydraError
from hydra_base.lib.objects import JSONObject, Dataset

import logging
log = logging.getLogger(__name__)

class TestScenario:

    def get_scenario(self, scenario_id, get_parent_data=False):
        """
            Utility function wrapper for a function that's called regularly
            Introduced as the JSONObject wrapper can be controlled more easily
        """
        return JSONObject(client.get_scenario(scenario_id, get_parent_data=get_parent_data))
    def get_scenario_data(self, scenario_id, get_parent_data=False):
        """
            Utility function wrapper for a function that's called regularly
            Introduced as the JSONObject wrapper can be controlled more easily
        """
        return client.get_scenario_data(scenario_id, get_parent_data=get_parent_data)

    def test_get_scenario(self, client, network_with_data):
        """
            Test to get the scenarios attached to a dataset
        """

        network = network_with_data

        #Create the new scenario
        scenario = network.scenarios[0]

        rs = scenario.resourcescenarios


        scenario_retrieved = client.get_scenario(scenario.id)
        assert len(scenario_retrieved.resourcescenarios) == len(rs)
        assert len(scenario_retrieved.resourcegroupitems) == len(scenario.resourcegroupitems)

        scenario_no_data = client.get_scenario(scenario.id, include_data=False)
        assert scenario_no_data.resourcescenarios == []
        assert len(scenario_no_data.resourcegroupitems) == len(scenario.resourcegroupitems)

        scenario_no_groupitems = client.get_scenario(scenario.id, include_group_items=False)
        assert len(scenario_no_groupitems.resourcescenarios) == len(scenario.resourcescenarios)
        assert scenario_no_groupitems.resourcegroupitems == []

    def test_get_scenario_by_name(self, client, network_with_data):
        """
            Test to get the scenarios attached to a dataset
        """

        network = network_with_data

        #Create the new scenario
        scenario = network.scenarios[0]

        rs = scenario.resourcescenarios


        scenario_retrieved = client.get_scenario_by_name(network.id, scenario.name)
        assert len(scenario_retrieved.resourcescenarios) == len(rs)
        assert len(scenario_retrieved.resourcegroupitems) == len(scenario.resourcegroupitems)

        scenario_no_data = client.get_scenario_by_name(network.id, scenario.name, include_data=False)
        assert scenario_no_data.resourcescenarios == []
        assert len(scenario_no_data.resourcegroupitems) == len(scenario.resourcegroupitems)

        scenario_no_groupitems = client.get_scenario_by_name(network.id, scenario.name, include_group_items=False)
        assert len(scenario_no_groupitems.resourcescenarios) == len(scenario.resourcescenarios)
        assert scenario_no_groupitems.resourcegroupitems == []

    def test_update(self, client, network_with_data):

        network =  network_with_data

        scenario = network.scenarios[0]
        scenario_id = scenario.id

        resource_scenario = scenario.resourcescenarios[0]
        resource_attr_id = resource_scenario.resource_attr_id

        dataset = Dataset({
            'type'      : 'descriptor',
            'name'      : 'Max Capacity',
            'unit_id'   : client.get_unit_by_abbreviation("m s^-1").id,
            'value'     : 'I am an updated test!',
        })

        new_resource_scenario = client.add_data_to_attribute(scenario_id, resource_attr_id, dataset)

        assert new_resource_scenario.dataset.value == 'I am an updated test!', "Value was not updated correctly!!"

    def test_add_scenario(self, client, network_with_data, dateformat):
        """
            Test adding a new scenario to a network.
        """
        network = network_with_data

        new_scenario = copy.deepcopy(network.scenarios[0])
        new_scenario.id = -1
        new_scenario.name = 'Scenario 2'
        new_scenario.description = 'Scenario 2 Description'
        start_time = datetime.datetime.now()
        end_time = start_time + datetime.timedelta(hours=10)
        new_scenario.start_time = start_time.strftime(dateformat)
        new_scenario.end_time = end_time.strftime(dateformat)
        new_scenario.time_step = "1 day"

        node_attrs = network.nodes[0].attributes

        #This is an example of 3 diffent kinds of data
        #A simple string (Descriptor)
        #A time series, where the value may be a 1-D array
        #A multi-dimensional array.
        descriptor = client.testutils.create_descriptor(client.testutils.get_by_name('node_attr_a', node_attrs), "new_descriptor")
        timeseries = client.testutils.create_timeseries(client.testutils.get_by_name('node_attr_b', node_attrs))

        for r in new_scenario.resourcescenarios:
            if r.resource_attr_id == client.testutils.get_by_name('node_attr_a', node_attrs).id:
                r.dataset = descriptor.dataset
            elif r.resource_attr_id == client.testutils.get_by_name('node_attr_b', node_attrs).id:
                r.dataset = timeseries.dataset

        scenario = JSONObject(client.add_scenario(network.id, new_scenario))

        assert scenario is not None
        assert len(scenario.resourcegroupitems) > 0
        assert len(scenario.resourcescenarios) > 0


    def test_update_scenario(self, client, network_with_data, dateformat):
        """
            Test updating an existing scenario.
        """
        network = network_with_data

        #Create the new scenario
        scenario = network.scenarios[0]
        scenario.name = 'Updated Scenario'
        scenario.description = 'Updated Scenario Description'
        start_time = datetime.datetime.now()
        end_time = start_time + datetime.timedelta(hours=10)
        scenario.start_time = start_time.strftime(dateformat)
        scenario.end_time = end_time.strftime(dateformat)
        scenario.time_step = "1 day"

        #Identify 2 nodes to play around with -- the first and last in the list.
        node1 = network.nodes[0]
        node2 = network.nodes[-1]

        #Identify 1 resource group item to edit (the last one in the list).
        item_to_edit = scenario.resourcegroupitems[-1]
        #Just checking that we're not changing an item that is already
        #assigned to this node..
        assert scenario.resourcegroupitems[-1].node_id != node2.id
        scenario.resourcegroupitems[-1].node_id   = node2.id

        descriptor = client.testutils.create_descriptor(
            client.testutils.get_by_name('node_attr_a', node1.attributes),
            "updated_descriptor")

        for resourcescenario in scenario.resourcescenarios:
            if resourcescenario.attr_id == descriptor.attr_id:
                resourcescenario.dataset = descriptor.dataset

        updated_scenario = JSONObject(client.update_scenario(scenario))

        assert updated_scenario is not None
        assert updated_scenario.id == scenario.id
        assert updated_scenario.name == scenario.name
        assert updated_scenario.description == scenario.description
        assert updated_scenario.start_time == scenario.start_time
        assert updated_scenario.end_time == scenario.end_time
        assert updated_scenario.time_step  == scenario.time_step
        assert len(updated_scenario.resourcegroupitems) > 0

        for i in updated_scenario.resourcegroupitems:
            if i.id == item_to_edit.id:
                assert i.node_id == node2.id
        assert len(updated_scenario.resourcescenarios) > 0

        for data in updated_scenario.resourcescenarios:
            if data.attr_id == descriptor['attr_id']:
                assert data.dataset.value == descriptor.dataset.value

    def test_get_dataset_scenarios(self, client, network_with_data):
        """
            Test to get the scenarios attached to a dataset
        """

        network = network_with_data

        #Create the new scenario
        scenario = network.scenarios[0]
        rs = scenario.resourcescenarios

        dataset_id_to_check = rs[0].dataset.id

        dataset_scenarios = [JSONObject(s) for s in client.get_dataset_scenarios(dataset_id_to_check)]

        assert len(dataset_scenarios) >= 1
        assert scenario.id in [s.id for s in dataset_scenarios]

        clone = client.clone_scenario(scenario.id)
        new_scenario = client.get_scenario(clone.id)

        dataset_scenarios = [JSONObject(s) for s in client.get_dataset_scenarios(dataset_id_to_check)]

        assert len(dataset_scenarios) >= 2

        assert set([scenario.id, new_scenario.id]).issubset(set([s.id for s in dataset_scenarios]))

    def test_update_resourcedata(self, client, network_with_data):
        """
            Test updating an existing scenario data.
            2 main points to test: 1: setting a value to null should remove
            the resource scenario
            2: changing the value should create a new dataset
        """
        network = network_with_data

        #Create the new scenario
        scenario = network.scenarios[0]
        num_old_rs = len(scenario.resourcescenarios)

        #Identify 2 nodes to play around with -- the first and last in the list.
        node1 = network.nodes[0]
        node1attr = client.testutils.get_by_name('node_attr_a', node1.attributes)
        node2 = network.nodes[-1]
        val_to_delete = client.testutils.get_by_name('node_attr_a', node2.attributes)

        descriptor = client.testutils.create_descriptor(node1attr,
                                                "updated_descriptor")

        rs_to_update = []
        updated_dataset_id = None
        scenario = client.get_scenario(scenario.id, include_data=True)
        for resourcescenario in scenario.resourcescenarios:
            ra_id = resourcescenario.resource_attr_id
            if ra_id == descriptor.resource_attr_id:
                updated_dataset_id = resourcescenario.dataset.id
                resourcescenario.dataset = descriptor.dataset
                rs_to_update.append(resourcescenario)
            elif ra_id == val_to_delete['id']:
                resourcescenario.dataset = None
                rs_to_update.append(resourcescenario)

        assert updated_dataset_id is not None

        new_resourcescenarios = [JSONObject(rs) for rs in client.update_resourcedata(scenario.id, rs_to_update)]

        assert len(new_resourcescenarios) == 1

        for rs in new_resourcescenarios:
            if rs.resource_attr_id == descriptor.resource_attr_id:
                assert rs.dataset.value == descriptor.dataset.value

        updated_scenario = client.get_scenario(scenario.id)

        num_new_rs = len(updated_scenario.resourcescenarios)

        assert num_new_rs == num_old_rs - 1


        for u_rs in updated_scenario.resourcescenarios:
            for rs in new_resourcescenarios:
                if u_rs.resource_attr_id == rs.resource_attr_id:
                    assert str(u_rs.dataset.value) == str(rs.dataset.value)
                    break

    def test_update_resourcedata_single_dataset_update_and_delete(self, client, network_with_data):
        """
            Test to ensure update_resourcedata does not update other
            datasets that it should not.
        """
        network = network_with_data

        scenario_1 = network.scenarios[0]
        scenario_2 = client.clone_scenario(scenario_1.id)
        scenario_2 = client.get_scenario(scenario_2.id)

        new_value = json.dumps({"index": {"2000-01-01":"test", "2000-02-01":"update"}})

        #Delete a timeseries from one scenario, so there's only 1 reference to that
        #dataset in tResourceSceanrio.
        ts_to_delete = []
        ra_id = None
        ts_id = None
        for rs in scenario_1.resourcescenarios:
            if rs.dataset.type == 'timeseries':
                ra_id = rs.resource_attr_id
                ts_id = rs.dataset.id
                rs.dataset = None
                ts_to_delete.append(rs)
                break

        ts_to_update = []
        for rs in scenario_2.resourcescenarios:
            if rs.resource_attr_id == ra_id:
                rs.dataset.value= new_value
                ts_to_update.append(rs)
                break

        client.update_resourcedata(scenario_1.id, ts_to_delete)

        client.update_resourcedata(scenario_2.id, ts_to_update)

        scenario_2_updated_1 = client.get_scenario(scenario_2.id)
        for rs in scenario_2_updated_1.resourcescenarios:
            if rs.resource_attr_id == ra_id:
                assert json.loads(rs.dataset.value) == json.loads(new_value)
                #Either the dataset is the same dataset, just updated or the dataset
                #has been removed and linked to a previous dataset, which must have a lower ID.
                assert rs.dataset.id <= ts_id
                break
        else:
            raise Exception("Couldn't find resource scenario. SOmething went wrong.")

    def test_update_resourcedata_consistency(self, client, network_with_data):
        """
            Test to ensure update_resourcedata does not update other
            datasets that it should not.
        """
        network = network_with_data

        scenario_1 = client.testutils.get_by_name("Scenario 1", network.scenarios)
        scenario_2 = client.clone_scenario(scenario_1.id)
        scenario_2 = client.get_scenario(scenario_2.id)

        #Identify 2 nodes to play around with -- the first and last in the list.
        node1 = network.nodes[0]


        descriptor = client.testutils.create_descriptor(client.testutils.get_by_name('node_attr_a', node1.attributes),
                                                "updated_descriptor")

        rs_to_update = self._get_rs_to_update(scenario_1, descriptor)


        #Update the value
        updated_data = client.update_resourcedata(scenario_1.id, rs_to_update)

        new_resourcescenarios = [JSONObject(rs) for rs in updated_data]

        rs_1_id = None
        updated_scenario_1 = client.get_scenario(scenario_1.id)
        for u_rs in updated_scenario_1.resourcescenarios:
            for rs in new_resourcescenarios:
                if u_rs.resource_attr_id == rs.resource_attr_id:
                    log.info(u_rs.dataset)
                    log.info(rs.dataset)
                    assert str(u_rs.dataset.value) == str(rs.dataset.value)
                    rs_1_id = u_rs.dataset
                    break

        scalar = client.testutils.create_descriptor(client.testutils.get_by_name('node_attr_a', node1.attributes), 200)

        rs_to_update = self._get_rs_to_update(scenario_2, scalar)

        new_resourcescenarios = client.update_resourcedata(scenario_2.id,
                                                           rs_to_update)
        rs_2_id = None
        #Check that scenario 2 has been updated correctly.
        updated_scenario_2 = client.get_scenario(scenario_2.id)
        for u_rs in updated_scenario_2.resourcescenarios:
            for rs in new_resourcescenarios:
                if u_rs.resource_attr_id == rs.resource_attr_id:
                    rs_2_id = u_rs.dataset.value
                    assert str(u_rs.dataset.value) == str(rs.dataset.value)
                    break
        log.critical("%s vs %s", rs_1_id, rs_2_id)
        #Check that this change has not affected scenario 1
        for u_rs in updated_scenario_1.resourcescenarios:
            for rs in new_resourcescenarios:
                if u_rs.resource_attr_id == rs.resource_attr_id:
                    assert str(u_rs.dataset.value) != str(rs.dataset.value)
                    break

    def _get_rs_to_update(self, scenario, rs):
        """
            Given a scenario, fetch all the RS which match the attribute ID
            of the rs passed in. These will be updated in an update call.
        """
        rs_to_update = []
        updated_dataset_id = None
        for resourcescenario in scenario.resourcescenarios:
            ra_id = resourcescenario.resource_attr_id
            if ra_id == rs['resource_attr_id']:
                updated_dataset_id = resourcescenario.dataset.id
                resourcescenario.dataset = rs.dataset
                rs_to_update.append(resourcescenario)

        assert updated_dataset_id is not None

        return rs_to_update

    def test_get_attributes_for_resource(self, client, network_with_data):
        """
            Test to check leng's questions about this not working correctly.
        """
        network = network_with_data

        #Create the new scenario
        scenario = network.scenarios[0]
        node1 = network.nodes[0]

        ra_to_update = client.testutils.get_by_name('node_attr_a', node1.attributes).id

        updated_val = None

        rs_to_update = []
        for resourcescenario in scenario.resourcescenarios:
            ra_id = resourcescenario.resource_attr_id
            if ra_id == ra_to_update:
                updated_val = resourcescenario.dataset.value
                resourcescenario.dataset.name = 'I am an updated dataset name'
                rs_to_update.append(resourcescenario)

        client.get_attributes_for_resource(network.id, scenario.id, 'NODE', [node1.id])

        client.update_resourcedata(scenario.id, rs_to_update)

        new_node_data = [JSONObject(d) for d in client.get_attributes_for_resource(network.id, scenario.id, 'NODE', [node1.id])]

        for new_val in new_node_data:
            if new_val.dataset.value == updated_val:
                assert new_val.dataset.name == 'I am an updated dataset name'

    def test_bulk_update_resourcedata(self, client, network_with_data):
        """
            Test updating scenario data in a number of scenarios at once.
            2 main points to test: 1: setting a value to null should remove
            the resource scenario
            2: changing the value should create a new dataset
        """
        network1 = network_with_data
        scenario1_to_update = network1.scenarios[0]
        clone = client.clone_scenario(network1.scenarios[0].id)
        scenario2_to_update = client.get_scenario(clone.id)

        #Identify 2 nodes to play around with -- the first and last in the list.
        node1 = network1.nodes[0]

        node1attr = client.testutils.get_by_name('node_attr_a', node1.attributes)

        node2 = network1.nodes[-1]

        descriptor = client.testutils.create_descriptor(node1attr, "updated_descriptor")

        val_to_delete = client.testutils.get_by_name('node_attr_a', node2.attributes)

        rs_to_update = []
        updated_dataset_id = None
        for resourcescenario in scenario1_to_update.resourcescenarios:
            ra_id = resourcescenario.resource_attr_id
            if ra_id == descriptor['resource_attr_id']:
                updated_dataset_id = resourcescenario.dataset.id
                resourcescenario.dataset = descriptor.dataset
                rs_to_update.append(resourcescenario)
            elif ra_id == val_to_delete['id']:
                resourcescenario.dataset = None
                rs_to_update.append(resourcescenario)

        assert updated_dataset_id is not None

        scenario_ids = []
        scenario_ids.append(scenario1_to_update.id)
        scenario_ids.append(scenario2_to_update.id)

        result = client.bulk_update_resourcedata(scenario_ids, rs_to_update)

        assert len(result) == 2

        updated_scenario1_data = client.get_scenario(scenario1_to_update.id)
        updated_scenario2_data = client.get_scenario(scenario2_to_update.id)

        for rs in updated_scenario1_data.resourcescenarios:
            ra_id = rs.resource_attr_id
            if ra_id == descriptor['resource_attr_id']:
                assert rs.dataset.value == descriptor.dataset.value
        for rs in updated_scenario2_data.resourcescenarios:
            ra_id = rs.resource_attr_id
            if ra_id == descriptor['resource_attr_id']:
                assert rs.dataset.value == descriptor.dataset.value



    def test_bulk_add_data(self, client, dateformat):

        data = []

        dataset1 = Dataset()

        dataset1.type = 'timeseries'
        dataset1.name = 'my time series'
        dataset1.unit_id = client.get_unit_by_abbreviation("ft^3").id


        t1 = datetime.datetime.now()
        t2 = t1+datetime.timedelta(hours=1)
        t3 = t1+datetime.timedelta(hours=2)

        t1 = t1.strftime(dateformat)
        t2 = t2.strftime(dateformat)
        t3 = t3.strftime(dateformat)

        val_1 = 1.234
        val_2 = 2.345
        val_3 = 3.456

        ts_val = json.dumps({0: {t1: val_1,
                      t2: val_2,
                      t3: val_3}})
        dataset1.value = ts_val
        data.append(dataset1)

        dataset2 = Dataset()
        dataset2.type = 'descriptor'
        dataset2.name = 'Max Capacity'
        dataset2.unit_id = client.get_unit_by_abbreviation("m s^-1").id

        dataset2.value ='I am an updated test!'

        data.append(dataset2)

        new_datasets = [Dataset(d) for d in client.bulk_insert_data(data)]

        assert len(new_datasets) == 2, "Data was not added correctly!"


    def test_bulk_insert_data(self, client, dateformat):
        import random
        from hydra_base.lib.data import bulk_insert_data

        num_datasets = 10
        datasets = []
        unit_id = client.get_unit_by_abbreviation("m s^-1").id
        for idx in range(num_datasets):
            ds = Dataset()
            ds.name = f"Bulk dataset {idx}"
            ds.type = "ARRAY"
            ds.unit_id = unit_id
            # Every third dataset is large
            data_sz = 10 if idx % 3 else 8192
            ds.value = [random.uniform(1, 100) for _ in range(data_sz)]

            datasets.append(ds)

        inserted = [Dataset(ds) for ds in client.bulk_insert_data(datasets)]
        assert len(inserted) == num_datasets, "Datasets insertion count mismatch"

        for ds in inserted:
            retrieved = client.get_dataset(ds.id)
            assert ds.value == retrieved.value


    def test_clone_scenario(self, client, network_with_data):

        network =  network_with_data

        assert len(network.scenarios) == 1, "The network should have only one scenario!"

        #self.create_constraint(network)

        network = JSONObject(client.get_network(network.id, include_data=True))

        scenario = network.scenarios[0]
        scenario_id = scenario.id

        clone = client.clone_scenario(scenario_id)
        clone1 = client.clone_scenario(scenario_id, scenario_name="My Cloned Scenario")
        clone2 = client.clone_scenario(scenario_id, retain_results=True)

        scenario_diff = JSONObject(client.compare_scenarios(scenario.id, clone.id))
        assert len(scenario_diff.resourcescenarios) == 0, "Scenarios should be the same but are not"

        #This should fail because there's already another scenario with this name
        with pytest.raises(HydraError):
            client.clone_scenario(scenario_id, scenario_name="My Cloned Scenario")
        #to deal with the broken flush caused by this exception


        new_scenario = client.get_scenario(clone.id)

        new_scenario_new_name = client.get_scenario(clone1.id)
        new_scenario_with_results = client.get_scenario(clone2.id)

        assert new_scenario_new_name.name == "My Cloned Scenario"

        assert len(new_scenario_with_results.resourcescenarios) > len(new_scenario.resourcescenarios)


        updated_network = JSONObject(client.get_network(new_scenario.network_id, include_data=True))


        assert len(updated_network.scenarios) == 4, "The network should have 4 scenarios!"

        assert len(updated_network.scenarios[1].resourcescenarios) > 0, "Data was not cloned!"

        #Find the correct scenarios to compare, to satisfy stupid postgres
        for original_scenario in network.scenarios:
            for updated_scenario in network.scenarios:
                if original_scenario.name == updated_scenario.name:
                    scen_2_vals = set([rs.dataset.value for rs in original_scenario.resourcescenarios])
                    scen_1_vals = set([rs.dataset.value for rs in updated_scenario.resourcescenarios])

                    assert scen_2_vals == scen_1_vals, "Data was not cloned correctly"
                    break


  #      scen_1_constraint  = network.scenarios[0].constraints.Constraint[0].value
        #scen_2_constraint  = updated_network.scenarios[1].constraints.Constraint[0].value
#
 #       assert scen_1_constraint == scen_2_constraint, "Constraints did not clone correctly!"

        scen_1_resourcegroupitems = network.scenarios[0].resourcegroupitems
        scen_2_resourcegroupitems = updated_network.scenarios[1].resourcegroupitems

        assert len(scen_1_resourcegroupitems) == len(scen_2_resourcegroupitems)

    def test_get_inherited_data(self, client, network_with_child_scenario):

        network = network_with_child_scenario

        sorted_scenario = sorted(network.scenarios, key=lambda x : x.id )

        parent = sorted_scenario[0]
        child  = sorted_scenario[1]

        assert len(child.resourcescenarios) == 0, "There's data in the child but there shouldn't be"

        #Check that the new scenario contains no data (as we've requested only its own data)
        retrieved_child_scenario = client.get_scenario(child.id)
        assert len(retrieved_child_scenario.resourcescenarios) == 0


        #Check that the new scenario contains all its data (that of its parent)
        retrieved_child_scenario = client.get_scenario(child.id, get_parent_data=True)
        assert len(retrieved_child_scenario.resourcescenarios) == len(parent.resourcescenarios)

    def test_get_third_level_inherited_data(self, client, network_with_grandchild_scenario):

        network = network_with_grandchild_scenario

        sorted_scenario = sorted(network.scenarios, key=lambda x : x.id )

        parent = sorted_scenario[0]
        child  = sorted_scenario[1]
        grandchild  = sorted_scenario[2]

        assert len(child.resourcescenarios) == 0, "There's data in the child but there shouldn't be"
        assert len(grandchild.resourcescenarios) == 0, "There's data in the child but there shouldn't be"

        #Check that the new scenario contains no data (as we've requested only its own data)
        retrieved_child_scenario = client.get_scenario(child.id)
        assert len(retrieved_child_scenario.resourcescenarios) == 0

        #Check that the grandchild scenario also contains no data (as we've requested only its own data)
        retrieved_grandchild_scenario = client.get_scenario(grandchild.id)
        assert len(retrieved_grandchild_scenario.resourcescenarios) == 0


        #Check that the grandchild scenario contains all its data (that of its grandparent)
        retrieved_grandchild_scenario = client.get_scenario(grandchild.id, get_parent_data=True)
        assert len(retrieved_grandchild_scenario.resourcescenarios) == len(parent.resourcescenarios)


        #do the same test for groups
        group_id = network.resourcegroups[0].id
        parent_group_items = client.get_resourcegroupitems(group_id, parent.id)
        assert len(parent_group_items) > 0
        grandchild_group_items = client.get_resourcegroupitems(group_id, grandchild.id)
        assert len(grandchild_group_items) == 0
        inherited_group_items = client.get_resourcegroupitems(group_id, grandchild.id, get_parent_items=True)
        assert len(inherited_group_items) == len(parent_group_items)



    def test_inherited_get_resource_scenario(self, client, network_with_child_scenario):

        network = network_with_child_scenario

        sorted_scenario = sorted(network.scenarios, key=lambda x : x.id )

        parent = sorted_scenario[0]
        child  = sorted_scenario[1]

        ra_to_query = parent.resourcescenarios[0].resource_attr_id

        #Request an RS without doing the parent lookup
        with pytest.raises(HydraError):
            rs = client.get_resource_scenario(ra_to_query, child.id)

        rs = client.get_resource_scenario(ra_to_query, child.id, get_parent_data=True)
        assert rs.scenario_id == parent.id

    def test_inherited_get_scenario_data(self, client, network_with_child_scenario):

        network = network_with_child_scenario

        sorted_scenario = sorted(network.scenarios, key=lambda x : x.id )

        parent = sorted_scenario[0]
        child  = sorted_scenario[1]

        child_scenario_data = client.get_scenario_data(child.id)
        assert len(child_scenario_data) == 0

        parent_scenario_data = client.get_scenario_data(parent.id)
        inherited_scenario_data = client.get_scenario_data(child.id, get_parent_data=True)

        assert len(inherited_scenario_data) == len(parent_scenario_data)

    def test_inherited_get_resource_data(self, client, network_with_child_scenario):

        network = network_with_child_scenario

        sorted_scenario = sorted(network.scenarios, key=lambda x : x.id )

        parent = sorted_scenario[0]
        child  = sorted_scenario[1]

        resource_to_query = network.nodes[0]

        child_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           child.id)

        assert len(child_resource_data) == 0

        parent_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           parent.id)

        inherited_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           child.id,
                                                           get_parent_data=True)

        assert len(inherited_resource_data) == len(parent_resource_data)

    def test_inherited_add_data_to_child(self, client, network_with_child_scenario):

        network = network_with_child_scenario

        sorted_scenario = sorted(network.scenarios, key=lambda x : x.id )

        parent = sorted_scenario[0]
        child  = sorted_scenario[1]

        #Specifically chose an attribute to update to remove non-determinisim
        #when trying to choose a resource to query later. Attribute 'node_attr_c' is a scalar, as
        #defined in the network creation. We use the name instead of an index or ID because
        #postgres ordering is non-deterministic.
        ra_to_update = None
        for a in network.nodes[0].attributes:
            if a.name == 'node_attr_a':
                ra_to_update = a.id

        #Update a value in the parent, leaving 1 different value between parent and child
        for rs in parent.resourcescenarios:
            if rs.resource_attr_id == ra_to_update:
                rs_to_update = rs
                break

        #Set the new value to 999
        rs_to_update.dataset.value = 999

        client.update_resourcedata(child.id,
                                        [rs_to_update])


        #Check that the new scenario contains all its data (that of its parent)
        new_scenario = client.get_scenario(child.id, get_parent_data=True)
        assert len(new_scenario.resourcescenarios) == len(parent.resourcescenarios)

        #Check that the new scenario contains no data (as we've requested only its own data)
        new_scenario = client.get_scenario(child.id)
        assert len(new_scenario.resourcescenarios) == 1

        child_scenario_data = client.get_scenario_data(child.id)
        assert len(child_scenario_data) == 1

        parent_scenario_data = client.get_scenario_data(parent.id)
        inherited_scenario_data = client.get_scenario_data(child.id, get_parent_data=True)

        #This returns 1 more for the inherited data because of the presence of
        #a new dataset in this result set (this function returns the UNIQUE) datasets
        #in each scenario.
        assert len(inherited_scenario_data) == len(parent_scenario_data)+1

        resource_to_query = network.nodes[0]

        child_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           child.id)

        assert len(child_resource_data) == 1

        parent_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           parent.id)

        inherited_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           child.id,
                                                           get_parent_data=True)

        assert len(inherited_resource_data) == len(parent_resource_data)


    def test_inherited_add_data_to_grandchild(self, client, network_with_grandchild_scenario):
        """
            Test inheritence by checking the get_scenario_data and get_scenario functions return the data
            of the grandchild and not the child, of the same RA
        """

        network = network_with_grandchild_scenario

        sorted_scenario = sorted(network.scenarios, key=lambda x : x.id )

        parent = sorted_scenario[0]
        child  = sorted_scenario[1]
        grandchild  = sorted_scenario[2]

        #Specifically chose an attribute to update to remove non-determinisim
        #when trying to choose a resource to query later. Attributea'node_attr_c' is a scalar, as
        #defined in the network creation. We use the name instead of an index or ID because
        #postgres ordering is non-deterministic.
        ra_to_update = None
        for a in network.nodes[0].attributes:
            if a.name == 'node_attr_a':
                ra_to_update = a.id

        #Update a value in the parent, leaving 1 different value between parent and child
        for rs in parent.resourcescenarios:
            if rs.resource_attr_id == ra_to_update:
                rs_to_update = rs
                break

        previous_rs_value = rs_to_update.dataset.value

        rs_to_update.dataset.value = 999

        client.update_resourcedata(child.id, [rs_to_update])

        rs_to_update.dataset.value = 1000

        client.update_resourcedata(grandchild.id, [rs_to_update])

        #check the child and grandchild have the same amount of data
        child_scenario = client.get_scenario(child.id, get_parent_data=True)
        grandchild_scenario = client.get_scenario(grandchild.id, get_parent_data=True)
        assert len(child_scenario.resourcescenarios) == len(grandchild_scenario.resourcescenarios)

        #find the RS in each scenario
        for rs in child_scenario.resourcescenarios:
            if rs.resource_attr_id == rs_to_update.resource_attr_id:
                assert int(rs.dataset.value) == 999
        #Check the grandchild has the correct value
        for rs in grandchild_scenario.resourcescenarios:
            if rs.resource_attr_id == rs_to_update.resource_attr_id:
                assert int(rs.dataset.value) == 1000


        #Check that the new scenario contains no data (as we've requested only its own data)
        child_scenario = client.get_scenario(child.id)
        grandchild_scenario = client.get_scenario(grandchild.id)
        assert len(child_scenario.resourcescenarios) == 1
        assert len(grandchild_scenario.resourcescenarios) == 1
        assert int(child_scenario.resourcescenarios[0].dataset.value) == 999
        assert int(grandchild_scenario.resourcescenarios[0].dataset.value) == 1000


        child_scenario_data = client.get_scenario_data(child.id)
        grandchild_scenario_data = client.get_scenario_data(grandchild.id)
        assert len(child_scenario_data) == 1
        assert len(grandchild_scenario_data) == 1

        parent_scenario_data = client.get_scenario_data(parent.id)
        child_scenario_data = client.get_scenario_data(child.id, get_parent_data=True)
        grandchild_scenario_data = client.get_scenario_data(grandchild.id, get_parent_data=True)

        #This returns 1 more for the inherited data because of the presence of
        #a new dataset in this result set (this function returns the UNIQUE) datasets
        #in each scenario.
        assert child_scenario_data != grandchild_scenario_data
        assert len(child_scenario_data) == len(parent_scenario_data)+1
        assert len(grandchild_scenario_data) == len(parent_scenario_data)+1

        resource_to_query = network.nodes[0]

        child_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           child.id)

        assert len(child_resource_data) == 1

        parent_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           parent.id)

        inherited_resource_data = client.get_resource_data('NODE',
                                                           resource_to_query.id,
                                                           child.id,
                                                           get_parent_data=True)

        assert len(inherited_resource_data) == len(parent_resource_data)


    def test_compare(self, client, network_with_data):

        network =  network_with_data


        assert len(network.scenarios) == 1, "The network should have only one scenario!"

        network = JSONObject(client.get_network(network.id))

        scenario = network.scenarios[0]
        scenario_id = scenario.id

        clone = client.clone_scenario(scenario_id, retain_results=True)
        new_scenario = client.get_scenario(clone.id)

        scenario_diff = JSONObject(client.compare_scenarios(scenario.id, new_scenario.id))

        assert len(scenario_diff.resourcescenarios) == 0, "Scenarios should be the same but are not"

        resource_scenario = new_scenario.resourcescenarios[0]
        resource_attr_id = resource_scenario.resource_attr_id

        dataset = Dataset({
            'type' : 'descriptor',
            'name' : 'Max Capacity',
            'unit_id' : client.get_unit_by_abbreviation("m s^-1").id,
            'value' : 'I am an updated test!',
        })

        client.add_data_to_attribute(scenario_id, resource_attr_id, dataset)

        item_to_remove = new_scenario.resourcegroupitems[0].id

        client.delete_resourcegroupitem(item_to_remove)

        updated_network = JSONObject(client.get_network(new_scenario.network_id))

        scenarios = updated_network.scenarios

        scenario_1 = None
        scenario_2 = None
        for s in scenarios:
            if s.id == new_scenario.id:
                scenario_1 = s
            else:
                scenario_2 = s

        scenario_diff = JSONObject(client.compare_scenarios(scenario_1.id, scenario_2.id))

        #print "Comparison result: %s"%(scenario_diff)

        assert len(scenario_diff.resourcescenarios) == 1, "Data comparison was not successful!"

     #   assert len(scenario_diff.constraints.common_constraints) == 1, "Constraint comparison was not successful!"

     #   assert len(scenario_diff.constraints.scenario_2_constraints) == 1, "Constraint comparison was not successful!"

        assert len(scenario_diff.groups.scenario_2_items) == 1, "Group comparison was not successful!"
        assert scenario_diff.groups.scenario_1_items == [], "Group comparison was not successful!"

    def test_purge_scenario(self, client, network_with_data):

        #Make a network with 2 scenarios
        client.clone_scenario(network_with_data.scenarios[0].id)
        net = client.get_network(network_with_data.id)

        scenarios_before = net.scenarios

        assert len(scenarios_before) == 2

        client.purge_scenario(scenarios_before[1].id)

        updated_net = JSONObject(client.get_network(net.id))

        scenarios_after = updated_net.scenarios

        assert len(scenarios_after) == 1

        with pytest.raises(HydraError):
            client.get_scenario(scenarios_before[1].id)

        assert scenarios_after[0].id == scenarios_before[0].id

    def test_delete_scenario(self, client, network_with_data):

        #Make a network with 2 scenarios
        client.clone_scenario(network_with_data.scenarios[0].id)
        net = client.get_network(network_with_data.id)

        scenarios_before = net.scenarios

        assert len(scenarios_before) == 2

        client.set_scenario_status(scenarios_before[1].id, 'X')

        updated_net = JSONObject(client.get_network(net.id))

        scenarios_after_delete = updated_net.scenarios

        assert len(scenarios_after_delete) == 1

        client.set_scenario_status(scenarios_before[1].id, 'A')

        updated_net2 = JSONObject(client.get_network(net.id))

        scenarios_after_reactivate = updated_net2.scenarios

        assert len(scenarios_after_reactivate) == 2

        client.set_scenario_status(scenarios_before[1].id, 'X')
        client.clean_up_network(net.id)
        updated_net3 = JSONObject(client.get_network(net.id))
        scenarios_after_cleanup = updated_net3.scenarios
        assert len(scenarios_after_cleanup) == 1
        with pytest.raises(HydraError):
            client.get_scenario(scenarios_before[1].id)


    def test_delete_resource_scenario(self, client, network_with_data):
        """
        """

        scenario = network_with_data.scenarios[0]

        before_rs = scenario.resourcescenarios

        rs_to_delete = before_rs[0]

        client.delete_resource_scenario(scenario.id, rs_to_delete.resource_attr_id, quiet=True)

        #test the quiet feature by trying to delete a non-existent RS
        with pytest.raises(HydraError):
            client.delete_resource_scenario(scenario.id, 999, quiet=False)

        client.delete_resource_scenario(scenario.id, 999, quiet=True)

        updated_scenario = client.get_scenario(scenario.id)

        assert len(updated_scenario.resourcescenarios) == len(before_rs)-1

    def test_delete_resource_scenarios(self, client, network_with_data):
        """
        """
        scenario = network_with_data.scenarios[0]

        before_rs = scenario.resourcescenarios

        rs_to_delete = [before_rs[0].resource_attr_id,before_rs[1].resource_attr_id]

        client.delete_resource_scenarios(scenario.id, rs_to_delete, quiet=True)

        #test the quiet feature by trying to delete a non-existent RS
        with pytest.raises(HydraError):
            client.delete_resource_scenarios(scenario.id, rs_to_delete, quiet=False)

        client.delete_resource_scenarios(scenario.id, rs_to_delete, quiet=True)

        updated_scenario = client.get_scenario(scenario.id)

        assert len(updated_scenario.resourcescenarios) == len(before_rs)-2

    def test_lock_scenario(self, client, network_with_data):

        network =  network_with_data

        network = client.get_network(network.id)

        scenario_to_lock = network.scenarios[0]
        scenario_id = scenario_to_lock.id

        log.info('Cloning scenario %s'%scenario_id)
        clone = client.clone_scenario(scenario_id)
        unlocked_scenario = client.get_scenario(clone.id)

        log.info("Locking scenario")
        client.lock_scenario(scenario_id)

        locked_scenario = client.get_scenario(scenario_id)

        assert locked_scenario.locked == 'Y'

        dataset = Dataset()

        dataset.type = 'descriptor'
        dataset.name = 'Max Capacity'
        dataset.unit_id = client.get_unit_by_abbreviation("m s^-1").id

        dataset.value = 'I am an updated test!'


        locked_resource_scenarios = []
        for rs in locked_scenario.resourcescenarios:
            if rs.dataset.type == 'descriptor':
                locked_resource_scenarios.append(rs)

        unlocked_resource_scenarios = []
        for rs in unlocked_scenario.resourcescenarios:
            if rs.dataset.type == 'descriptor':
                unlocked_resource_scenarios.append(rs)

        resource_attr_id = unlocked_resource_scenarios[0].resource_attr_id

        locked_resource_scenarios_value = None
        for rs in locked_scenario.resourcescenarios:
            if rs.resource_attr_id == resource_attr_id:
                locked_resource_scenarios_value = rs.dataset

        unlocked_resource_scenarios_value = None
        for rs in unlocked_scenario.resourcescenarios:
            if rs.resource_attr_id == resource_attr_id:
                unlocked_resource_scenarios_value = rs.dataset
        log.info("Updating a shared dataset")
        ds = unlocked_resource_scenarios_value

        updated_ds = JSONObject(client.update_dataset(ds.id, ds.name, ds.type, ds.value, ds.unit_id, ds.metadata))

        updated_unlocked_scenario = client.get_scenario(unlocked_scenario.id)

        updated_locked_scenario = client.get_scenario(locked_scenario.id)

        locked_resource_scenarios_value = None
        for rs in updated_locked_scenario.resourcescenarios:
            if rs.resource_attr_id == resource_attr_id:
                locked_resource_scenarios_value = rs.datset

        unlocked_resource_scenarios_value = None
        for rs in updated_unlocked_scenario.resourcescenarios:
            if rs.resource_attr_id == resource_attr_id:
                unlocked_resource_scenarios_value = rs.dataset

        with pytest.raises(HydraError):
            client.add_data_to_attribute(scenario_id, resource_attr_id, dataset)

        #THe most complicated situation is this:
        #Change a dataset in an unlocked scenario, which is shared by a locked scenario.
        #The original dataset should stay connected to the locked scenario and a new
        #dataset should be created for the edited scenario.
        client.add_data_to_attribute(unlocked_scenario.id, resource_attr_id, dataset)

        updated_unlocked_scenario = client.get_scenario(unlocked_scenario.id)

        #This should not have changed
        updated_locked_scenario = client.get_scenario(locked_scenario.id)

        locked_resource_scenarios_value = None
        for rs in updated_locked_scenario.resourcescenarios:
            if rs.resource_attr_id == resource_attr_id:
                locked_resource_scenarios_value = rs.dataset

        unlocked_resource_scenarios_value = None
        for rs in updated_unlocked_scenario.resourcescenarios:
            if rs.resource_attr_id == resource_attr_id:
                unlocked_resource_scenarios_value = rs.dataset

        assert locked_resource_scenarios_value.hash != unlocked_resource_scenarios_value.hash

        item_to_remove = locked_scenario.resourcegroupitems[0].id

        with pytest.raises(HydraError):
            client.delete_resourcegroupitem(item_to_remove)

        log.info("Locking scenario")
        client.unlock_scenario(scenario_id)

        locked_scenario = client.get_scenario(scenario_id)

        assert locked_scenario.locked == 'N'


    def test_get_attribute_data(self, client, network_with_data):
        """
            Test for retrieval of data for an attribute in a scenario.
        """

        new_net = network_with_data

        s = new_net.scenarios[0]

        nodes = new_net.nodes

        resource_attr = client.testutils.get_by_name('node_attr_a', nodes[0].attributes)

        attr_id = resource_attr.attr_id

        all_matching_ras = []
        for n in nodes:
            for ra in n.attributes:
                if ra.attr_id == attr_id:
                    all_matching_ras.append(ra)
                    continue

        retrieved_resource_scenarios = client.get_attribute_datasets(attr_id, s.id)

        rs_dict = {}
        for rs in retrieved_resource_scenarios:
            rs_dict[rs.resource_attr_id] = rs

        assert len(retrieved_resource_scenarios) == len(all_matching_ras)

        for rs in s.resourcescenarios:
            if rs_dict.get(rs.resource_attr_id):
                matching_rs = rs_dict[rs.resource_attr_id]
                assert str(rs.dataset.hash) == str(matching_rs.dataset.hash)

    def test_copy_data_from_scenario(self, client, network_with_data):

        """
            Test copy_data_from_scenario : test that one scenario
            can be updated to contain the data of another with the same
            resource attrs.
        """

        network = network_with_data


        network = client.get_network(network.id)

        scenario = network.scenarios[0]
        source_scenario_id = scenario.id

        clone = client.clone_scenario(source_scenario_id, retain_results=True)
        cloned_scenario = client.get_scenario(clone.id)

        resource_scenario = cloned_scenario.resourcescenarios[0]
        resource_attr_id = resource_scenario.resource_attr_id

        dataset = Dataset()
        dataset.type = 'descriptor'
        dataset.name = 'Max Capacity'
        dataset.unit_id = client.get_unit_by_abbreviation("m s^-1").id

        dataset.value = 'I am an updated test!'

        client.add_data_to_attribute(source_scenario_id, resource_attr_id, dataset)

        scenario_diff = JSONObject(client.compare_scenarios(source_scenario_id, cloned_scenario.id))

        assert len(scenario_diff.resourcescenarios) == 1, "Data comparison was not successful!"

        client.copy_data_from_scenario([resource_attr_id], cloned_scenario.id, source_scenario_id)

        scenario_diff = JSONObject(client.compare_scenarios(source_scenario_id, cloned_scenario.id))

        assert len(scenario_diff.resourcescenarios) == 0, "Scenario update was not successful!"

    def test_merge_scenario(self, client, networkmaker):

        """
            Test merge_scenario : test that one scenario
            can be updated to contain the data of another with the same
            resource attrs.
        """

        sourcenetwork = networkmaker.create()
        targetnetwork = networkmaker.create()

        source_scenario = sourcenetwork.scenarios[0]
        target_scenario = targetnetwork.scenarios[0]
        source_scenario_id = source_scenario.id

        resource_scenario = source_scenario.resourcescenarios[0]
        resource_attr_id = resource_scenario.resource_attr_id

        dataset = Dataset()
        dataset.type = 'descriptor'
        dataset.name = 'Max Capacity'
        dataset.unit_id = client.get_unit_by_abbreviation("m s^-1").id
        dataset.value = 'I am an updated test!'

        client.add_data_to_attribute(source_scenario_id, resource_attr_id, dataset)

        source_scenario = client.get_scenario(source_scenario.id)
        target_scenario = client.get_scenario(target_scenario.id)
        source_rs = sorted(source_scenario.resourcescenarios, key=lambda x: x.resource_attr_id)
        target_rs = sorted(target_scenario.resourcescenarios, key=lambda x: x.resource_attr_id)
        differences = []
        for i, rs in enumerate(source_rs):
            if rs.dataset.value != target_rs[i].dataset.value:
                differences.append(rs.id)
        assert len(differences) > 0

        client.merge_scenarios(source_scenario.id, target_scenario.id)

        #Now check that the scenarios are the same by making sure all the datasets
        #are the same
        updated_target_network = client.get_network(targetnetwork.id)
        merged_scenario = client.get_scenario(sorted(updated_target_network.scenarios, key=lambda x: x.cr_date)[-1].id)
        source_scenario = client.get_scenario(source_scenario.id)
        source_rs = sorted(source_scenario.resourcescenarios, key=lambda x: x.resource_attr_id)
        merged_rs = sorted(merged_scenario.resourcescenarios, key=lambda x: x.resource_attr_id)
        differences = []
        for i, rs in enumerate(source_rs):
            if rs.dataset.value != merged_rs[i].dataset.value:
                differences.append(rs.id)
        assert len(differences) == 0


    def test_set_resourcescenario_dataset(self, client, network_with_data):

        """
            Test the direct setting of a dataset id on a resource scenario
        """

        network = network_with_data


        network = client.get_network(network.id)

        scenario = network.scenarios[0]
        source_scenario_id = scenario.id

        clone = client.clone_scenario(source_scenario_id)
        cloned_scenario = client.get_scenario(clone.id)

        resource_scenario = cloned_scenario.resourcescenarios[0]
        resource_attr_id = resource_scenario.resource_attr_id

        dataset = Dataset()
        dataset.type = 'descriptor'
        dataset.name = 'Max Capacity'
        dataset.unit_id = client.get_unit_by_abbreviation("m s^-1").id

        dataset.value = 'I am an updated test!'

        new_ds = client.add_dataset(dataset.type, dataset.value, dataset.unit_id, {}, dataset.name, flush=True)

        client.set_rs_dataset(resource_attr_id, source_scenario_id, new_ds.id)

        updated_net = client.get_network(network.id)

        updated_scenario = updated_net.scenarios[0]
        scenario_rs = updated_scenario.resourcescenarios
        for rs in scenario_rs:
            if rs.resource_attr_id == resource_attr_id:
                assert rs.dataset.value == 'I am an updated test!'

    def test_add_data_to_attribute(self, client, network_with_data):

        network = network_with_data

        empty_ra = network.links[0].attributes[-1]

        scenario = network.scenarios[0]
        scenario_id = scenario.id

        resource_scenario = scenario.resourcescenarios[0]
        resource_attr_id = resource_scenario.resource_attr_id

        dataset = Dataset()
        dataset.type = 'descriptor'
        dataset.name = 'Max Capacity'
        dataset.unit_id = client.get_unit_by_abbreviation("m s^-1").id

        dataset.value = 'I am an updated test!'

        updated_resource_scenario = client.add_data_to_attribute(scenario_id, resource_attr_id, dataset)

        new_resource_scenario = client.add_data_to_attribute(scenario_id, empty_ra.id, dataset)

        assert updated_resource_scenario.dataset.value == 'I am an updated test!', "Value was not updated correctly!!"
        assert new_resource_scenario.dataset.value == 'I am an updated test!', "Value was not updated correctly!!"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import os
from ..exceptions import HydraError
from ..import config
import base64
import logging
log = logging.getLogger(__name__)

def add_image(name, file,**kwargs):
    path = config.get('filesys', 'img_src')
    try:
        os.makedirs(path)
    except OSError:
        pass

    path = os.path.join(path, name)

    #The safest way to check if a file exists is to try to open
    #it. If the open succeeds, then throw an exception to this effect.
    try:
        f = open(path)
        raise HydraError("A file with this name (%s) already exists!"%(name))
    except IOError:
        pass


    log.info("Path: %r" % path)
    if not path.startswith(path):
        log.critical("Could not open file: %s"%name)
        return False

    f = open(path, 'wb') # if this fails, the client will see an
    # # internal error.

    try:
        for data in file:
            f.write(data)
        f.close()

        log.debug("File written: %r" % name)

    except:
        log.critical("Error writing to file: %s", name)
        f.close()
        os.remove(name)
        log.debug("File removed: %r" % name)
        return False

    return True

def get_image(name,**kwargs):
    path = config.get('filesys', 'img_src')

    path = os.path.join(path, name)

    #The safest way to check if a file exists is to try to open
    #it. If the open succeeds, then throw an exception to this effect.
    try:
        f = open(path, 'rb')
    except IOError:
        raise HydraError("File with name (%s) does not exist!"%(name))

    #read the contents of the file
    imageFile = f.read()

    #encode the contents of the file as a byte array
    #encodedFile = base64.b64encode(imageFile)

    return imageFile

def remove_image(name,**kwargs):
    path = config.get('filesys', 'img_src')

    path = os.path.join(path, name)
    if(os.path.exists(path)):
        os.remove(path)
    else:
        raise HydraError("File with name (%s) does not exist!"%(name))

    return True


def add_file(resource_type, resource_id, name, file,**kwargs):
    path = config.get('filesys', 'file_src')
    path = os.path.join(path, resource_type)
    try:
        os.makedirs(path)
    except OSError:
        pass

    path = os.path.join(path, str(resource_id))
    try:
        os.makedirs(path)
    except OSError:
        pass

    path = os.path.join(path, name)

    #The safest way to check if a file exists is to try to open
    #it. If the open succeeds, then throw an exception to this effect.
    try:
        f = open(path)
        raise HydraError("A file with this name (%s) already exists!"%(name))
    except IOError:
        pass


    log.info("Path: %r" % path)
    if not path.startswith(path):
        log.critical("Could not open file: %s"%name)
        return False

    f = open(path, 'wb') # if this fails, the client will see an
    # # internal error.

    try:
        for data in file:
            f.write(data)

        log.debug("File written: %r" % name)
        f.close()

    except:
        log.critical("Error writing to file: %s", name)
        f.close()
        os.remove(name)
        log.debug("File removed: %r" % name)
        return False

    return True

def get_file(resource_type, resource_id, name,**kwargs):
    path = config.get('filesys', 'file_src')

    path = os.path.join(path, resource_type, str(resource_id), name)

    #The safest way to check if a file exists is to try to open
    #it. If the open succeeds, then throw an exception to this effect.
    try:
        f = open(path, 'rb')
    except IOError:
        raise HydraError("File with name (%s) does not exist!"%(name))

    #read the contents of the file
    file_to_send = f.read()
    f.close()

    #encode the contents of the file as a byte array

    #encodedFile = base64.b64encode(file_to_send)

    return file_to_send

def remove_file(resource_type, resource_id, name,**kwargs):
    path = config.get('filesys', 'file_src')

    path = os.path.join(path, resource_type, str(resource_id), name)

    if(os.path.exists(path)):
        os.remove(path)
    else:
        raise HydraError("File with name (%s) does not exist!"%(name))

    return True




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import logging
from decimal import Decimal

from operator import mul
from ..exceptions import HydraError, ValidationError
import numpy as np
import pandas as pd
import re
from functools import reduce
import json

from hydra_base.util.hydra_dateutil import get_datetime
log = logging.getLogger(__name__)

def json_to_df(json_dataframe):
    """
     Create a pandas dataframe from a json string.
     Pandas does not maintain the order of the index; it sorts it.
     To manage this, we avail of python 3.6+ inhernet dict ordering to identify
     the correct index order, and then reindex the dataframe to this after
     it has been created
    """

    #Load the json dataframe into a native dict
    data_dict = json.loads(json_dataframe)
    
    #load the json dataframe into a pandas dataframe
    df = pd.read_json(json_dataframe, convert_axes=False)

    #extraxt the ordered index from the dict
    ordered_index = list(data_dict[df.columns[0]].keys())

    #extract the order of columns
    ordered_cols = list(data_dict.keys())
    
    #Make the df index and columns a string so it is comparable to the dict index (which must be string based)
    df.index = df.index.astype(str)
    df.columns = df.columns.astype(str)
    
    #set the column ordering as per the incoming json data
    df = df[ordered_cols]

    #reindex the dataframe to have the correct order
    df = df.reindex(ordered_index)
    
    return df

def array_dim(arr):
    """Return the size of a multidimansional array.
    """
    dim = []
    while True:
        try:
            dim.append(len(arr))
            arr = arr[0]
        except TypeError:
            return dim

def check_array_struct(array):
    """
        Check to ensure arrays are symmetrical, for example:
        [[1, 2, 3], [1, 2]] is invalid
    """

    #If a list is transformed into a numpy array and the sub elements
    #of this array are still lists, then numpy failed to fully convert
    #the list, meaning it is not symmetrical.
    try:
        arr = np.array(array)
    except:
        raise HydraError("Array %s is not valid."%(array,))
    if type(arr[0]) is list:
        raise HydraError("Array %s is not valid."%(array,))


def arr_to_vector(arr):
    """Reshape a multidimensional array to a vector.
    """
    dim = array_dim(arr)
    tmp_arr = []
    for n in range(len(dim) - 1):
        for inner in arr:
            for i in inner:
                tmp_arr.append(i)
        arr = tmp_arr
        tmp_arr = []
    return arr


def vector_to_arr(vec, dim):
    """Reshape a vector to a multidimensional array with dimensions 'dim'.
    """
    if len(dim) <= 1:
        return vec
    array = vec
    while len(dim) > 1:
        i = 0
        outer_array = []
        for m in range(reduce(mul, dim[0:-1])):
            inner_array = []
            for n in range(dim[-1]):
                inner_array.append(array[i])
                i += 1
            outer_array.append(inner_array)
        array = outer_array
        dim = dim[0:-1]

        return array

def _get_val(val, full=False):
    """
        Get the value(s) of a dataset as a single value or as 1-d list of
        values. In the special case of timeseries, when a check is for time-based
        criteria, you can return the entire timeseries.
    """

    try:
        val = val.strip()
    except:
        pass

    logging.debug("%s, type=%s", val, type(val))

    if isinstance(val, float):
        return val

    if isinstance(val, int):
        return val


    if isinstance(val, np.ndarray):
        return list(val)

    try:
        val = float(val)
        return val
    except:
        pass

    try:
        val = int(val)
        return val
    except:
        pass

    if type(val) == pd.DataFrame:

        if full:
            return val

        newval = []
        values = val.values
        for v in values:
            newv = _get_val(v)
            if type(newv) == list:
                newval.extend(newv)
            else:
                newval.append(newv)
        val = newval

    elif type(val) == dict:

        if full:
            return val

        newval = []
        for v in val.values():
            newv = _get_val(v)
            if type(newv) == list:
                newval.extend(newv)
            else:
                newval.append(newv)
        val = newval

    elif type(val) == list or type(val) == np.ndarray:
        newval = []
        for arr_val in val:
            v = _get_val(arr_val)
            newval.append(v)
        val = newval
    return val

def get_restriction_as_dict(restriction_xml):
    """
    turn:
    ::

            <restrictions>
                <restriction>
                    <type>MAXLEN</type>
                    <value>3</value>
                </restriction>
                <restriction>
                    <type>VALUERANGE</type>
                    <value><item>1</item><item>10</item></value>
                </restriction>
            </restrictions>

    into:
    ::

        {
            'MAXLEN' : 3,
            'VALUERANGE' : [1, 10]
        }

    """
    restriction_dict = {}
    if restriction_xml is None:
        return restriction_dict

    if restriction_xml.find('restriction') is not None:
        restrictions = restriction_xml.findall('restriction')
        for restriction in restrictions:
            restriction_type = restriction.find('type').text
            restriction_val  = restriction.find('value')
            val = None
            if restriction_val is not None:
                if restriction_val.text.strip() != "":
                    val = _get_val(restriction_val.text)
                else:
                    items = restriction_val.findall('item')
                    val = []
                    for item in items:
                        val.append(_get_val(item.text))
            restriction_dict[restriction_type] = val
    return restriction_dict


def validate_ENUM(in_value, restriction):
    """
        Test to ensure that the given value is contained in the provided list.
        the value parameter must be either a single value or a 1-dimensional list.
        All the values in this list must satisfy the ENUM
    """
    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_ENUM(subval, restriction)
    else:
        if value not in restriction:
            raise ValidationError("ENUM : %s"%(restriction))


def validate_BOOL(in_value, restriction):
    """
        Validation for the generic Boolean type consisting of arbitrary
        'true' and 'false' values.
    """
    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_BOOL(subval, restriction)
    else:
        if value not in restriction:
            raise ValidationError("BOOL")


def validate_BOOLYN(in_value, restriction):
    """
        Restriction is not used here. It is just present to be
        in line with all the other validation functions

    """
    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_BOOLYN(subval, restriction)
    else:
        if value not in ('Y', 'N'):
            raise ValidationError("BOOLYN")


def validate_BOOL10(value, restriction):
    value = _get_val(value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_BOOL10(subval, restriction)
    else:
        if value not in (1, 0):
            raise ValidationError("BOOL10")

def validate_NUMPLACES(in_value, restriction):
    """
        the value parameter must be either a single value or a 1-dimensional list.
        All the values in this list must satisfy the condition
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_NUMPLACES(subval, restriction)
    else:
        restriction = int(restriction) # Just in case..
        dec_val = Decimal(str(value))
        num_places = dec_val.as_tuple().exponent * -1 #exponent returns a negative num
        if restriction != num_places:
            raise ValidationError("NUMPLACES: %s"%(restriction))

def validate_VALUERANGE(in_value, restriction):
    """
        Test to ensure that a value sits between a lower and upper bound.
        Parameters: A Decimal value and a tuple, containing a lower and upper bound,
        both as Decimal values.
    """
    if len(restriction) != 2:
        raise ValidationError("Template ERROR: Only two values can be specified in a date range.")
    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_VALUERANGE(subval, restriction)
    else:
        min_val = Decimal(restriction[0])
        max_val = Decimal(restriction[1])
        val     = Decimal(value)
        if val < min_val or val > max_val:
            raise ValidationError("VALUERANGE: %s, %s"%(min_val, max_val))

def validate_DATERANGE(value, restriction):
    """
        Test to ensure that the times in a timeseries fall between a lower and upper bound
        Parameters: A timeseries in the form [(datetime, val), (datetime, val)..]
        and a tuple containing the lower and upper bound as datetime objects.
    """
    if len(restriction) != 2:
        raise ValidationError("Template ERROR: Only two values can be specified in a date range.")

    if type(value) == pd.DataFrame:
        dates = [get_datetime(v) for v in list(value.index)]
    else:
        dates = value

    if type(dates) is list:
        for date in dates:
            validate_DATERANGE(date, restriction)
        return

    min_date = get_datetime(restriction[0])
    max_date = get_datetime(restriction[1])
    if value < min_date or value > max_date:
        raise ValidationError("DATERANGE: %s <%s> %s"%(min_date,value,max_date))

def validate_MAXLEN(value, restriction):
    """
        Test to ensure that a list has the prescribed length.
        Parameters: A list and an integer, which defines the required length of
        the list.
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]
    else:
        return

    if len(value) > restriction:
        raise ValidationError("MAXLEN: %s"%(restriction))

def validate_NOTNULL(value, restriction):
    """
        Restriction is not used here. It is just present to be
        in line with all the other validation functions

    """

    if value is None or str(value).lower == 'null':
        raise ValidationError("NOTNULL")

def validate_ISNULL(value, restriction):
    """
        Restriction is not used here. It is just present to be
        in line with all the other validation functions

    """

    if value is not None and str(value).lower != 'null':
        raise ValidationError("ISNULL")

def validate_EQUALTO(in_value, restriction):
    """
        Test to ensure that a value is equal to a prescribed value.
        Parameter: Two values, which will be compared for equality.
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_EQUALTO(subval, restriction)
    else:
        if value != restriction:
            raise ValidationError("EQUALTO: %s"%(restriction))

def validate_NOTEQUALTO(in_value, restriction):
    """
        Test to ensure that a value is NOT equal to a prescribed value.
        Parameter: Two values, which will be compared for non-equality.
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_NOTEQUALTO(subval, restriction)
    else:
        if value == restriction:
            raise ValidationError("NOTEQUALTO: %s"%(restriction))

def validate_LESSTHAN(in_value, restriction):
    """
        Test to ensure that a value is less than a prescribed value.
        Parameter: Two values, which will be compared for the difference..
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_LESSTHAN(subval, restriction)
    else:
        try:
            if value >= restriction:
                raise ValidationError("LESSTHAN: %s"%(restriction))
        except TypeError:
            # Incompatible types for comparison.
            raise ValidationError("LESSTHAN: Incompatible types %s"%(restriction))


def validate_LESSTHANEQ(value, restriction):
    """
        Test to ensure that a value is less than or equal to a prescribed value.
        Parameter: Two values, which will be compared for the difference..
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_LESSTHANEQ(subval, restriction)
    else:
        try:
            if value > restriction:
                raise ValidationError("LESSTHANEQ: %s" % (restriction))
        except TypeError:
            # Incompatible types for comparison.
            raise ValidationError("LESSTHANEQ: Incompatible types %s"%(restriction))


def validate_GREATERTHAN(in_value, restriction):
    """
        Test to ensure that a value is greater than a prescribed value.
        Parameter: Two values, which will be compared for the difference..
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_GREATERTHAN(subval, restriction)
    else:
        try:
            if value <= restriction:
                raise ValidationError("GREATERTHAN: %s" % (restriction))
        except TypeError:
            # Incompatible types for comparison.
            raise ValidationError("GREATERTHAN: Incompatible types %s"%(restriction))

def validate_GREATERTHANEQ(value, restriction):
    """
        Test to ensure that a value is greater than or equal to a prescribed value.
        Parameter: Two values, which will be compared for the difference..
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_GREATERTHANEQ(subval, restriction)
    else:
        try:
            if value < restriction:
                raise ValidationError("GREATERTHANEQ: %s" % (restriction))
        except TypeError:
            # Incompatible types for comparison.
            raise ValidationError("GREATERTHANEQ: Incompatible types %s"%(restriction))

def validate_MULTIPLEOF(in_value, restriction):
    """
        Test to ensure that a value is a multiple of a specified restriction value.
        Parameters: Numeric value and an integer
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(in_value)
    if type(value) is list:
        for subval in value:
            if type(subval) is tuple:
                subval = subval[1]
            validate_MULTIPLEOF(subval, restriction)
    else:
        try:
            if value % restriction != 0:
                raise ValidationError("MULTIPLEOF: %s" % (restriction))
        except TypeError:
            # Incompatible types for comparison.
            raise ValidationError("MULTIPLEOF: Incompatible types %s"%(restriction))

def validate_SUMTO(in_value, restriction):
    """
        Test to ensure the values of a list sum to a specified value:
        Parameters: a list of numeric values and a target to which the values
        in the list must sum
    """
    #Sometimes restriction values can accidentally be put in the template <item>100</items>,
    #Making them a list, not a number. Rather than blowing up, just get value 1 from the list.
    if type(restriction) is list:
        restriction = restriction[0]

    value = _get_val(in_value, full=True)

    if len(value) == 0:
        return

    flat_list = _flatten_value(value)

    try:
        sum(flat_list)
    except:
        raise ValidationError("List cannot be summed: %s"%(flat_list,))

    if sum(flat_list) != restriction:
        raise ValidationError("SUMTO: %s"%(restriction))

def validate_INCREASING(in_value, restriction):
    """
        Test to ensure the values in a list are increasing.
        Parameters: a list of values and None. The none is there simply
        to conform with the rest of the validation routines.
    """

    flat_list = _flatten_value(in_value)

    previous = None
    for a in flat_list:
        if previous is None:
            previous = a
            continue
        try:
            if a < previous:
                raise ValidationError("INCREASING")
        except TypeError:
            raise  ValueError("INCREASING: Incompatible types")
        previous = a

def validate_DECREASING(in_value,restriction):
    """
        Test to ensure the values in a list are decreasing.
        Parameters: a list of values and None. The none is there simply
        to conform with the rest of the validation routines.
    """
    flat_list = _flatten_value(in_value)

    previous = None
    for a in flat_list:
        if previous is None:
            previous = a
            continue
        try:
            if a > previous:
                raise ValidationError("DECREASING")
        except TypeError:
            raise  ValueError("DECREASING: Incompatible types")
        previous = a

def validate_EQUALTIMESTEPS(value, restriction):
    """
        Ensure that the timesteps in a timeseries are equal. If a restriction
        is provided, they must be equal to the specified restriction.

        Value is a pandas dataframe.
    """
    if len(value) == 0:
        return

    if type(value) == pd.DataFrame:
        if str(value.index[0]).startswith('9999'):
            tmp_val = value.to_json().replace('9999', '1900')
            value = pd.read_json(tmp_val)


    #If the timeseries is not datetime-based, check for a consistent timestep
    if type(value.index) == pd.Int64Index:
        timesteps = list(value.index)
        timestep = timesteps[1] - timesteps[0]
        for i, t in enumerate(timesteps[1:]):
            if timesteps[i] - timesteps[i-1] != timestep:
                raise ValidationError("Timesteps not equal: %s"%(list(value.index)))


    if not hasattr(value.index, 'inferred_freq'):
        raise ValidationError("Timesteps not equal: %s"%(list(value.index),))

    if restriction is None:
        if value.index.inferred_freq is None:
            raise ValidationError("Timesteps not equal: %s"%(list(value.index),))
    else:
        if value.index.inferred_freq != restriction:
            raise ValidationError("Timesteps not equal: %s"%(list(value.index),))

validation_func_map = dict(
    ENUM = validate_ENUM,
    BOOL = validate_BOOL,
    BOOLYN = validate_BOOLYN,
    BOOL10 = validate_BOOL10,
    NUMPLACES = validate_NUMPLACES,
    VALUERANGE = validate_VALUERANGE,
    DATERANGE = validate_DATERANGE,
    MAXLEN = validate_MAXLEN,
    EQUALTO = validate_EQUALTO,
    NOTEQUALTO = validate_NOTEQUALTO,
    LESSTHAN = validate_LESSTHAN,
    LESSTHANEQ = validate_LESSTHANEQ,
    GREATERTHAN = validate_GREATERTHAN,
    GREATERTHANEQ = validate_GREATERTHANEQ,
    MULTIPLEOF = validate_MULTIPLEOF,
    SUMTO = validate_SUMTO,
    INCREASING = validate_INCREASING,
    DECREASING = validate_DECREASING,
    EQUALTIMESTEPS = validate_EQUALTIMESTEPS,
    NOTNULL        = validate_NOTNULL,
    ISNULL         = validate_ISNULL,
)

def validate_value(restriction_dict, inval):
    if len(restriction_dict) == 0:
        return

    try:
        for restriction_type, restriction in restriction_dict.items():
            func = validation_func_map.get(restriction_type)
            if func is None:
                raise Exception("Validation type {} does not exist".format(restriction_type,))
            func(inval, restriction)
    except ValidationError as e:
        log.exception(e)
        err_val = re.sub('\s+', ' ', str(inval)).strip()
        if len(err_val) > 60:
            err_val = "{}...".format(err_val[:60])
        raise HydraError("Validation error ({}). Val {} does not conform with rule {}".format(restriction_type, err_val, e.args[0]))
    except Exception as e:
        log.exception(e)
        raise HydraError("An error occurred in validation. ({})".format(e))

def _flatten_value(value):
    """
        1: Turn a multi-dimensional array into a 1-dimensional array
        2: Turn a timeseries of values into a single 1-dimensional array
    """

    if type(value) == pd.DataFrame:
        value = value.values.tolist()

    if type(value) != list:
        raise ValidationError("Value %s cannot be processed."%(value))

    if len(value) == 0:
        return

    flat_list = _flatten_list(value)

    return flat_list

def _flatten_list(l):
    flat_list = []
    for item in l:
        if type(item) is list:
            flat_list.extend(_flatten_list(item))
        else:
            flat_list.append(item)
    return flat_list

if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import logging
log = logging.getLogger(__name__)

from decimal import Decimal
import pandas as pd

import json
import six
from .. import config

from collections import namedtuple

# Python 2 and 3 compatible string checking
try:
    basestring
except NameError:
    basestring = str

def count_levels(value):
    """
        Count how many levels are in a dict:
        scalar, list etc = 0
        {} = 0
        {'a':1} = 1
        {'a' : {'b' : 1}} = 2
        etc...
    """
    if not isinstance(value, dict) or len(value) == 0:
        return 0
    elif len(value) == 0:
        return 0 #An emptu dict has 0
    else:
        nextval = list(value.values())[0]
        return 1 + count_levels(nextval)

def flatten_dict(value, target_depth=1, depth=None):
    """
        Take a hashtable with multiple nested dicts and return a
        dict where the keys are a concatenation of each sub-key.

        The depth of the returned array is dictated by target_depth, defaulting to 1

        ex: {'a' : {'b':1, 'c': 2}} ==> {'a_b': 1, 'a_c': 2}

        Assumes a constant structure actoss all sub-dicts. i.e. there isn't
        one sub-dict with values that are both numbers and sub-dicts.
    """

    #failsafe in case someone specified null
    if target_depth is None:
        target_depth = 1

    values = list(value.values())
    if len(values) == 0:
        return {}
    else:
        if depth is None:
            depth = count_levels(value)

        if isinstance(values[0], dict) and len(values[0]) > 0:
            subval = list(values[0].values())[0]
            if not isinstance(subval, dict) != 'object':
                return value

            if target_depth >= depth:
                return value

            flatval = {}
            for k in value.keys():
                subval = flatten_dict(value[k], target_depth, depth-1)
                for k1 in subval.keys():
                    flatval[str(k)+"_"+str(k1)] = subval[k1];
            return flatval
        else:
            return value

def generate_data_hash(dataset_dict):

    d = dataset_dict
    if d.get('metadata') is None:
        d['metadata'] = {}

    hash_string = "%s %s %s %s"%(
                                str(d['unit_id']),
                                str(d['type']).lower(),
                                d['value'],
                                d['metadata'])

    log.debug("Generating data hash from: %s", hash_string)

    hash_data = hash(hash_string)

    log.debug("Data hash: %s", hash_data)

    return hash_data

def get_val(dataset, timestamp=None):
    """
        Turn the string value of a dataset into an appropriate
        value, be it a decimal value, array or time series.

        If a timestamp is passed to this function,
        return the values appropriate to the requested times.

        If the timestamp is *before* the start of the timeseries data, return None
        If the timestamp is *after* the end of the timeseries data, return the last
        value.

        The raw flag indicates whether timeseries should be returned raw -- exactly
        as they are in the DB (a timeseries being a list of timeseries data objects,
        for example) or as a single python dictionary

    """

    val = dataset.get_value()

    if dataset.type == 'array':
        #TODO: design a mechansim to retrieve this data if it's stored externally
        return json.loads(val)

    elif dataset.type == 'descriptor':
        return str(val)
    elif dataset.type == 'scalar':
        return Decimal(str(val))
    elif dataset.type == 'timeseries':
        #TODO: design a mechansim to retrieve this data if it's stored externally

        seasonal_year = config.get('DEFAULT','seasonal_year', '1678')
        seasonal_key = config.get('DEFAULT', 'seasonal_key', '9999')
        val = val.replace(seasonal_key, seasonal_year)

        timeseries = pd.read_json(val, convert_axes=True)

        if isinstance(timeseries.index, pd.DatetimeIndex):
            if timeseries.index.tz is None:
                timeseries = timeseries.tz_localize('UTC')
            else:
                timeseries = timeseries.tz_convert('UTC')


        if timestamp is None:
            return timeseries
        else:
            try:
                idx = timeseries.index
                #Seasonal timeseries are stored in the year
                #1678 (the lowest year pandas allows for valid times).
                #Therefore if the timeseries is seasonal,
                #the request must be a seasonal request, not a
                #standard request

                if type(idx) == pd.DatetimeIndex:
                    if set(idx.year) == set([int(seasonal_year)]):
                        if isinstance(timestamp,  list):
                            seasonal_timestamp = []
                            for t in timestamp:
                                t_1900 = t.replace(year=int(seasonal_year))
                                seasonal_timestamp.append(t_1900)
                            timestamp = seasonal_timestamp
                        else:
                            timestamp = [timestamp.replace(year=int(seasonal_year))]

                pandas_ts = timeseries.reindex(timestamp, method='ffill')

                #If there are no values at all, just return None
                if len(pandas_ts.dropna()) == 0:
                    return None

                #Replace all numpy NAN values with None
                pandas_ts = pandas_ts.where(pandas_ts.notnull(), None)

                val_is_array = False
                if len(pandas_ts.columns) > 1:
                    val_is_array = True

                if val_is_array:
                    if type(timestamp) is list and len(timestamp) == 1:
                        ret_val = pandas_ts.loc[timestamp[0]].values.tolist()
                    else:
                        ret_val = pandas_ts.loc[timestamp].values.tolist()
                else:
                    col_name = pandas_ts.loc[timestamp].columns[0]
                    if type(timestamp) is list and len(timestamp) == 1:
                        ret_val = pandas_ts.loc[timestamp[0]].loc[col_name]
                    else:
                        ret_val = pandas_ts.loc[timestamp][col_name].values.tolist()

                return ret_val

            except Exception as e:
                log.critical("Unable to retrive data. Check timestamps.")
                log.critical(e)

def get_json_as_string(json_string_or_dict):
    """
        Take a dict or string and return a string.
        The dict will be json dumped.
        The string will json parsed to check for json validity. In order to deal
        with strings which have been json encoded multiple times, keep json decoding
        until a dict is retrieved or until a non-json structure is identified.
    """

    if isinstance(json_string_or_dict, dict):
        return json.dumps(json_string_or_dict)

    if(isinstance(json_string_or_dict, six.string_types)):
        try:
            return get_json_as_string(json.loads(json_string_or_dict))
        except:
            return json_string_or_dict

def get_json_as_dict(json_string_or_dict):
    """
        Take a dict or string and return a dict if the data is json-encoded.
        The string will json parsed to check for json validity. In order to deal
        with strings which have been json encoded multiple times, keep json decoding
        until a dict is retrieved or until a non-json structure is identified.
    """

    if isinstance(json_string_or_dict, dict):
        return json_string_or_dict

    if(isinstance(json_string_or_dict, six.string_types)):
        try:
            return get_json_as_dict(json.loads(json_string_or_dict))
        except:
            return json_string_or_dict


class NullAdapterType(type):
    """
      Metaclass for NullAdapter. Overriding __getattr__ here
      enables class and static attrs to be simulated.
    """
    def __getattr__(self, *args, **kwargs):
        return self


class NullAdapter(metaclass=NullAdapterType):
    """
      A class that does nothing.
      This may be used in place of any adapter in order to disable
      the functionality of that adapter type without requiring other
      code to be changed.
      From the caller's perspective, all attributes exist on this class,
      including nested attributes, and all are callable.
    """
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return self

    def __getattr__(self, *args, **kwargs):
        return self

    def __getitem__(self, idx):
        return None

    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration



from sqlalchemy.exc import NoResultFound
import numbers

from hydra_base.db import get_session
from hydra_base.exceptions import HydraError
from .mongostorageadapter import MongoStorageAdapter
from .datasetmanager import DatasetManager

import logging
log = logging.getLogger(__name__)


class MongoDatasetManager(DatasetManager):
    """
    Implements the descriptor protocol to manage accesses to Dataset._value
    and route these to the correct storage provider
    """
    def __init__(self, ref_key="value_ref", loc_key=None):
        mongo_config = MongoStorageAdapter.get_mongo_config()
        self.ref_key = ref_key
        self.loc_key = loc_key if loc_key else mongo_config["value_location_key"]
        self.threshold = mongo_config["threshold"]
        self.loc_mongo_direct = mongo_config["direct_location_token"]
        self.mongo = MongoStorageAdapter()  # Default config from hydra.ini


    def __set_name__(self, dataset, attr):
        """ Always '_value' """
        self.instattr = attr


    def __get__(self, dataset, dtype=None):
        value = getattr(dataset, self.ref_key)
        log.debug(f"* Dataset read: on {dataset=} {dtype=}")
        if loc := self.get_storage_location(dataset):
            log.debug(f"* External storage {loc=} with id='{value}'")
            if loc == self.loc_mongo_direct:
                return self.mongo.get_document_by_object_id(value)["value"]

        return value


    def __set__(self, dataset, value):
        if not value or isinstance(value, numbers.Number):
            """ Empty or numeric value """
            size = 0
        else:
            try:
                size = len(value)
            except TypeError as err:
                raise HydraError(f"{value=} written to dataset has invalid type {type(value)=}") from err

        log.debug(f"* Dataset write: {size=} {value=} on {dataset=}")

        loc = self.get_storage_location(dataset)
        is_mongo_direct = loc == self.loc_mongo_direct

        if is_mongo_direct:
            """ Already in external storage """
            if size <= self.threshold:
                """ Value has shrunk so restore to SQL DB """
                self.delete_storage_location(dataset)
                oid = getattr(dataset, self.ref_key)
                self.mongo.delete_document_by_object_id(oid)
                setattr(dataset, self.ref_key, value)
                log.debug(f"Deleted {oid=} on {dataset.id=} and restored {value=} to DB")
            else:
                """ Update in external storage """
                oid = getattr(dataset, self.ref_key)
                self.mongo.set_document_value(oid, value)
        elif size > self.threshold:
            """ Create in external storage """
            _id = self.mongo.insert_document(value)
            self.set_storage_location(dataset, self.loc_mongo_direct)
            setattr(dataset, self.ref_key, str(_id))
            log.debug(f"* External create in {self.loc_mongo_direct=} as {_id=}")
        else:
            """ In SQL DB: set value directly """
            log.debug(f"* Direct set {value=}")
            setattr(dataset, self.ref_key, value)


    def get_storage_location(self, dataset):
        if not dataset:
            return
        for datum in dataset.metadata:
            if datum.key == self.loc_key:
                return datum.value


    def delete_storage_location(self, dataset):
        for idx, datum in enumerate(dataset.metadata):
            if datum.key == self.loc_key:
                break
        dataset.metadata.pop(idx)
        get_session().delete(datum)


    def set_storage_location(self, dataset, location):
        from hydra_base.db.model import Metadata
        m = Metadata(key=self.loc_key, value=location)
        dataset.metadata.append(m)



import fsspec
import h5py
import inspect
import logging
import os
import s3fs

from botocore.exceptions import ClientError
from urllib.parse import urlparse
from functools import wraps

from hydra_base import config
from hydra_base.util import NullAdapter
from hydra_base.lib.storage.readers import group_reader_map

log = logging.getLogger(__name__)


def filestore_url(argname, rewrite_func="url_to_filestore_path"):
    """
      Decorator which rewrites the <argname> argument to the decorated function
      to include the appropriate path according to the <rewrite_func> argument.
    """
    def dfunc(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            fas = inspect.getfullargspec(func)
            idx = fas.args.index(argname)
            av = inspect.getargvalues(inspect.currentframe())
            """
              NB. self appears in args from getfullargspec, but not in
              getargvalues().locals(). Adjust the index to account for
              this difference.
            """
            idx -= 1
            pre_arg = av.locals["args"][idx]
            try:
                rw_func = getattr(self, rewrite_func)
            except AttributeError as ae:
                raise ValueError(f"Invalid rewrite_func '{rewrite_func}' for @filestore_url") from ae
            # rw_func is a bound method so may be called directly
            post_arg = rw_func(pre_arg)
            insert = (*av.locals["args"][:idx], post_arg, *av.locals["args"][idx+1:])
            return func(self, *insert, **kwargs)

        return wrapper
    return dfunc


class HdfStorageAdapter():
    """
       Utilities to describe and retrieve data from HDF storage
    """

    def __init__(self):
        self.fsspec_args = dict(
            mode='rb',
            anon=self._get_anon(),
            default_fill_cache=True
        )
        self.config = self.__class__.get_hdf_config()
        self.filestore_path = self.config.get("hdf_filestore")
        if self.filestore_path and not os.path.exists(self.filestore_path):
            os.makedirs(self.filestore_path, exist_ok=True)

    def _get_anon(self):
        """
            If there are AWS credentials present in the environment, 
            then assume that this user will be not anonymous
        """

        self.accesskeyid = os.getenv('AWS_ACCESS_KEY_ID')
        self.secretaccesskey = os.getenv('AWS_SECRET_ACCESS_KEY')
        if self.accesskeyid not in ('' , None) and self.secretaccesskey not in ('', None):
            return False

        home = os.getenv('HOME')
        default_credentials_path = os.path.join(home, '.aws', 'credentials')
        credentials_path = os.getenv('AWS_SHARED_CREDENTIALS_FILE', default_credentials_path)
        #If there is a credentials file, then assume the user is not anonymous
        if os.path.exists(credentials_path):
            return False

        return True

    @staticmethod
    def get_hdf_config(config_key="storage_hdf", **kwargs):
        numeric = ()
        boolean = ("disable_hdf", )
        hdf_keys = [k for k in config.CONFIG.options(config_key) if k not in config.CONFIG.defaults()]
        hdf_items = {k: config.CONFIG.get(config_key, k) for k in hdf_keys}
        for k in numeric:
            hdf_items[k] = int(hdf_items[k])
        for k in boolean:
            hdf_items[k] = hdf_items[k].lower() in ("true", "yes")

        return hdf_items

    @filestore_url("url")
    def open_hdf_url(self, url, **kwargs):
        try:
            with fsspec.open(url, **self.fsspec_args) as fp:
                return h5py.File(fp.fs.open(url), mode='r')
        except (ClientError, FileNotFoundError, PermissionError) as e:
            raise ValueError(f"Unable to access url: {url}") from e

    def url_to_filestore_path(self, url, do_raise=True, check_exists=False, **kwargs):
        u = urlparse(url)
        if u.path == "" and do_raise:
            raise ValueError(f"Invalid URL: {url}")
        if self.filestore_path and u.scheme in ("", "file"):
            relpath = u.path.lstrip("/")
            relurl = os.path.join(self.filestore_path, relpath)
        elif u.scheme == "path":
            relurl = u.path
        else:
            relurl = url

        return relurl

    def file_exists_at_url(self, url, **kwargs):
        try:
            url = self.url_to_filestore_path(url)
            with fsspec.open(url, **self.fsspec_args) as fp:
                return fp.fs.exists(url)
        except (ValueError, FileNotFoundError, PermissionError):
            return False

    @filestore_url("url")
    def get_group_info(self, url, groupname):
        """
          Returns a dict containing two keys:
            - index: a dict of 'name', 'length', 'dtype' for
                     the index of the <group> arg
            - series: an array of dicts, each containing 'name',
                      'length', 'dtype' for each column of the
                      <group> arg
        """
        reader = self.make_group_reader(url, groupname)
        index_info = reader.get_index_info()
        columns = reader.get_columns_of_group()
        series_info = [reader.get_series_info(c) for c in columns]

        return {
          "index": index_info,
          "series": series_info
        }

    @filestore_url("url")
    def get_index_info(self, url, groupname):
        reader = self.make_group_reader(url, groupname)
        return reader.get_index_info()

    @filestore_url("url")
    def get_index_range(self, url, groupname, start=0, end=None):
        reader = self.make_group_reader(url, groupname)
        return reader.get_index_range(start, end)

    @filestore_url("url")
    def get_series_info(self, url, groupname=None, columns=None):
        reader = self.make_group_reader(url, groupname)
        if isinstance(columns, str):
            """
              Assume any str arg represents a single column
              name which should have been passed as a single-
              element Sequence, *unless that string is empty*
              in which case it is equivalent to an empty
              container and represents *all* columns
            """
            columns = (columns,) if len(columns) > 0 else None
        if not columns:
            columns = reader.get_columns_of_group()
        return [reader.get_series_info(c) for c in columns]

    @filestore_url("url")
    def file_size(self, url, **kwargs):
        with fsspec.open(url, **self.fsspec_args) as fp:
            size_bytes = fp.fs.size(fp.path)
        return size_bytes

    @filestore_url("url")
    def get_hdf_groups(self, url, **kwargs):
        h5f = self.open_hdf_url(url)
        return [*h5f.keys()]

    @filestore_url("url")
    def get_group_columns(self, url, groupname):
        reader = self.make_group_reader(url, groupname)
        return reader.get_columns_of_group()

    @filestore_url("url")
    def get_group_index(self, url, groupname):
        reader = self.make_group_reader(url, groupname)
        return reader.get_index_range(start=0, end=None)

    @filestore_url("url")
    def get_group_shape(self, url, groupname):
        reader = self.make_group_reader(url, groupname)
        return reader.get_group_shape()

    def get_columns_as_dataframe(self, url, groupname=None, columns=None, start=None, end=None, **kwargs):
        json_opts = {"date_format": "iso"}

        reader = self.make_group_reader(url, groupname)
        if isinstance(columns, str):
            """
              Assume any str arg represents a single column
              name which should have been passed as a single-
              element Sequence, *unless that string is empty*
              in which case it is equivalent to an empty
              container and represents *all* columns
            """
            columns = (columns,) if len(columns) > 0 else None
        if not columns:
            columns = reader.get_columns_of_group()

        df = reader.get_columns_as_dataframe(columns, start=start, end=end)

        return df.to_json(**json_opts)

    def equivalent_local_path(self, url, **kwargs):
        u = urlparse(url)
        filesrc = f"{u.netloc}{u.path}"
        if u.scheme == "path":
            return os.path.dirname(filesrc), os.path.basename(filesrc)
        filedir = os.path.dirname(filesrc)

        if self.filestore_path is not None:
            destdir = os.path.join(self.filestore_path, filedir.strip('/'))
        else:
            destdir = filedir
        return destdir, os.path.basename(filesrc)

    def retrieve_s3_file(self, url, **kwargs):
        u = urlparse(url)
        if not u.scheme == "s3":
            return
        filesrc = f"{u.netloc}{u.path}"
        destdir, filename = self.equivalent_local_path(url)

        if not os.path.exists(destdir):
            try:
                os.makedirs(destdir)
            except OSError as err:
                raise OSError(f"Unable to create local path at {destdir}: {err}")

        destfile = os.path.join(destdir, filename)
        fs = s3fs.S3FileSystem(anon=self._get_anon())
        log.info(f"Retrieving {url} to {destfile} ...")
        fs.get(filesrc, destfile)
        file_sz = os.stat(destfile).st_size
        log.info(f"Retrieved {destfile} ({file_sz} bytes)")
        return destfile, file_sz

    def list_local_files(self):
        import glob
        files = {}
        pattern = os.path.join(self.filestore_path, "**")
        for p in glob.iglob(pattern, recursive=True):
            if not os.path.isfile(p):
                continue
            files[p] = os.stat(p).st_size
        return files

    def purge_local_file(self, filename):
        """
          This prevents directory traversal by:
            - relative path components
            - ~user path components
            - $ENV_VAR components
            - paths containing hard or symbolic links

          A valid target file must be all of:
            - a real absolute filesystem path
            - a subtree of the filestore
            - not a directory
            - not a link
            - not a device file or pipe
            - owned by the Hydra user

          In addition, the filestore_path may not be:
            - undefined
            - the root filesystem
            - the root of any mount point

          ValueError is raised if any of these conditions
          are not met.
        """
        real_fsp = os.path.realpath(self.filestore_path)
        if not self.filestore_path or real_fsp == '/' or os.path.ismount(real_fsp):
            raise ValueError(f"Invalid filestore configuration value '{self.filestore_path}'")

        expanded = os.path.expandvars(filename)
        if expanded != filename:
            raise ValueError(f"Invalid path '{filename}': Arguments may not contain variables")
        target = os.path.realpath(expanded)
        if os.path.commonprefix([target, self.filestore_path]) != self.filestore_path:
            raise ValueError(f"Invalid path '{filename}': Only filestore files may be purged")

        if not os.path.exists(target):
            raise ValueError(f"Invalid path '{filename}': File does not exist")

        # Tests for directories, device files and pipes, and existence again
        if not os.path.isfile(target):
            raise ValueError(f"Invalid path '{filename}': Only regular files may be purged")

        if os.getuid() != os.stat(target).st_uid:
            raise ValueError(f"Invalid path '{filename}': File is not owned by "
                             f"user {os.getlogin()} ({os.getuid()})")
        try:
            os.unlink(target)
        except OSError as oe:
            raise ValueError(f"Invalid path '{filename}': Unable to purge file") from oe

        return target

    def make_group_reader(self, url, groupname):
        """
        Returns an instance of the appropriate GroupReader subclass for the <groupname>
        argument in the file at <url>.
        The required type is first looked up by get_group_reader() and an instance of
        this is returned.
        """
        hf = self.open_hdf_url(url)
        if not groupname:
            # Use first group in file
            try:
                groupname = [*hf.keys()][0]
            except IndexError as ie:
                raise ValueError(f"Data source {url} contains no groups") from ie
        Reader = self.get_group_reader(url, groupname)
        return Reader(hf, groupname)

    def get_group_reader(self, url, groupname):
        """
        Returns the type of the appropriate GroupReader subclass for the <groupname>
        argument in the file at <url>.
        The internal format of the group used by Pandas is first identified by
        identify_group_format(), and a reader capable of handling this type is then
        looked up in the group_reader_map.
        """
        group_type = self.identify_group_format(url, groupname)
        try:
            Reader = group_reader_map[group_type]
        except KeyError:
            # Not-None group_type was returned, but we don't have a reader for it
            raise ValueError(f"No reader available for group of type {group_type}")

        return Reader

    def identify_group_format(self, url, groupname):
        """
        Returns a string identifying the Pandas format of the group <groupname> in
        the file at <url>.
        This string, stored as the `pandas_type` attribute on the HDF group, indicates
        the layout of datasets and indices for the group and therefore the type of
        GroupReader required.
        """
        hf = self.open_hdf_url(url)
        try:
            group = hf[groupname]
        except KeyError as ke:
            raise ValueError(f"File at {url} contains no group {groupname}") from ke

        try:
            pandas_type = group.attrs["pandas_type"]
        except KeyError as ke:
            raise ValueError(f"File at {url} has invalid format") from ke

        return pandas_type.decode()


"""
  If config defines [hdf_storage]::disable_hdf as "True",
  the above HdfStorageAdapter is replaced with a Null object.
"""
if HdfStorageAdapter.get_hdf_config().get("disable_hdf"):
    HdfStorageAdapter = NullAdapter



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import json
import logging
import bcrypt
from collections import defaultdict
import logging

from sqlalchemy import Column,\
ForeignKey,\
text,\
Integer,\
String,\
LargeBinary,\
TIMESTAMP,\
BIGINT,\
SMALLINT,\
Float,\
Text, \
JSON, \
DateTime,\
Unicode

from sqlalchemy.orm import relationship, backref, noload, joinedload
from sqlalchemy import inspect, func
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import case
from sqlalchemy import UniqueConstraint, and_, or_
from sqlalchemy.dialects import mysql

from hydra_base.lib.objects import JSONObject, Dataset as JSONDataset
from hydra_base.lib.cache import cache
from hydra_base.exceptions import HydraError, PermissionError, ResourceNotFoundError
from hydra_base.db import DeclarativeBase as Base, get_session
from hydra_base.util import generate_data_hash, get_val, get_json_as_string
from hydra_base import config

log = logging.getLogger(__name__)

# Python 2 and 3 compatible string checking
# TODO remove this when Python2 support is dropped.
try:
    basestring
except NameError:
    basestring = str

def get_user_id_from_engine(ctx):
    """
        The session will have a user ID bound to it when checking for the permission.
    """
    if hasattr(get_session(), 'user_id'):
        return get_session().user_id
    else:
        return None

def _is_admin(user_id):
    """
        Is the specified user an admin
    """
    from .permissions import User

    user = get_session().query(User).filter(User.id==user_id).one()

    if user.is_admin():
        return True
    else:
        return False

class Inspect(object):
    _parents = []
    _children = []

    def get_columns_and_relationships(self):
        return inspect(self).attrs.keys()

class AuditMixin(object):

    cr_date = Column(TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    @declared_attr
    def created_by(cls):
        return Column(Integer, ForeignKey('tUser.id'), default=get_user_id_from_engine)

    @declared_attr
    def updated_by(cls):
        return Column(Integer, ForeignKey('tUser.id'), onupdate=get_user_id_from_engine)

    updated_at = Column(DateTime, nullable=False, server_default=text(u'CURRENT_TIMESTAMP'), onupdate=func.current_timestamp())

class PermissionControlled(object):
    def set_owner(self, user_id, read='Y', write='Y', share='Y'):
        owner = None
        for o in self.owners:
            if user_id == o.user_id:
                owner = o
                break
        else:
            owner = self.__ownerclass__()
            setattr(owner, self.__ownerfk__, self.id)
            owner.user_id = int(user_id)
            self.owners.append(owner)

        if read is not None:
            owner.view = read
        if write is not None:
            owner.edit = write
        if share is not None:
            owner.share = share

        return owner

    def unset_owner(self, user_id):
        owner = None
        if str(user_id) == str(self.created_by):
            log.warning("Cannot unset %s as owner, as they created the dataset", user_id)
            return
        for o in self.owners:
            if user_id == o.user_id:
                owner = o
                get_session().delete(owner)
                break

    def _is_open(self):
        """
            Check to see whether this entity is visible globally, without any
            need for permission checking
        """
        pass

    def check_read_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can read this dataset
        """
        if str(user_id) == str(self.created_by):
            return True

        if is_admin is None:
            is_admin = _is_admin(user_id)

        if is_admin is True:
            return True


        #Check if this entity is publicly open, therefore no need to check permissions.
        if self._is_open() == True:
            return True

        for owner in self.owners:
            if int(owner.user_id) == int(user_id):
                if owner.view == 'Y':
                    break
        else:
            if do_raise is True:
                raise PermissionError("Permission denied. User %s does not have read"
                             " access on %s '%s' (id=%s)" %
                             (user_id, self.__class__.__name__, self.name, self.id))
            else:
                return False
        #Check that the user is in a group which can read this network
        self.check_group_read_permission(user_id)

        return True

    def check_group_read_permission(self, user_id):
        """
            1: Find which user groups a user is if
            2: Check if any of these groups has permission to read the object
        """

    def check_write_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can write this dataset
        """
        if str(user_id) == str(self.created_by):
            return True

        if is_admin is None:
            is_admin = _is_admin(user_id)

        if is_admin is True:
            return True

        for owner in self.owners:
            if owner.user_id == int(user_id):
                if owner.view == 'Y' and owner.edit == 'Y':
                    break
        else:
            if do_raise is True:
                raise PermissionError("Permission denied. User %s does not have edit"
                             " access on %s '%s' (id=%s)" %
                             (user_id, self.__class__.__name__, self.name, self.id))
            else:
                return False

        return True

    def check_share_permission(self, user_id, is_admin=None):
        """
            Check whether this user can write this dataset
        """

        if str(user_id) == str(self.created_by):
            return True

        if is_admin is None:
            is_admin = _is_admin(user_id)

        if is_admin is True:
            return True

        for owner in self.owners:
            if owner.user_id == int(user_id):
                if owner.view == 'Y' and owner.share == 'Y':
                    return True
        else:
            raise PermissionError("Permission denied. User %s does not have share"
                             " access on %s '%s' (id=%s)" %
                             (user_id, self.__class__.__name__, self.name, self.id))




"""resource_type_id

Revision ID: f52ccd2b7ffb
Revises: a48ad41579de
Create Date: 2018-04-17 15:01:35.411693

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f52ccd2b7ffb'
down_revision = 'a48ad41579de'
branch_labels = None
depends_on = None

import logging
log = logging.getLogger(__name__)


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tResourceType

        try:
            op.alter_column('tResourceType', 'resource_type_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tResourceType
        try:
            op.drop_table('tResourceType_new')
        except:
            log.info("tResourceType_new isn't there")

        try:
            # ## tResourceType
            op.create_table(
                'tResourceType_new',
                    sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('type_id', sa.Integer(), sa.ForeignKey('tTemplateType.id'), primary_key=False, nullable=False),
                    sa.Column('ref_key', sa.String(60)),
                    sa.Column('network_id', sa.Integer(),  sa.ForeignKey('tNetwork.id'), nullable=True),
                    sa.Column('node_id', sa.Integer(),  sa.ForeignKey('tNode.id'), nullable=True),
                    sa.Column('link_id', sa.Integer(),  sa.ForeignKey('tLink.id'), nullable=True),
                    sa.Column('group_id', sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), nullable=True),
                    sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                    sa.UniqueConstraint('network_id', 'type_id', name='net_type_1'),
                    sa.UniqueConstraint('node_id', 'type_id', name='node_type_1'),
                    sa.UniqueConstraint('link_id', 'type_id',  name = 'link_type_1'),
                    sa.UniqueConstraint('group_id', 'type_id', name = 'group_type_1'),
            )

            op.execute("insert into tResourceType_new (id, type_id, ref_key, network_id, node_id, link_id, group_id, cr_date) select resource_type_id, type_id, ref_key, network_id, node_id, link_id, group_id, cr_date from tResourceType")

            op.rename_table('tResourceType','tResourceType_old')
            op.rename_table('tResourceType_new', 'tResourceType')
            op.drop_table('tResourceType_old')

        except Exception as e:
            log.exception(e)




def downgrade():
    pass
